##############################################################################
# Copyright 2016-2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""Schema definition of a Experiment, a collection of ExperimentSetting objects and a main program body (or ansatz).

This schema is widely useful for defining and executing many common types of algorithms / applications, including state
and process tomography, and the variational quantum eigensolver.
"""

import json
import logging
import warnings
from collections.abc import Generator, Mapping, Sequence
from json import JSONEncoder
from typing import (
    Any,
    Callable,
    Optional,
    Union,
    cast,
)

from pyquil.experiment._calibration import CalibrationMethod
from pyquil.experiment._memory import (
    pauli_term_to_measurement_memory_map,
    pauli_term_to_preparation_memory_map,
)
from pyquil.experiment._program import (
    measure_qubits,
    parameterized_readout_symmetrization,
    parameterized_single_qubit_measurement_basis,
    parameterized_single_qubit_state_preparation,
)
from pyquil.experiment._result import ExperimentResult
from pyquil.experiment._setting import ExperimentSetting, TensorProductState, _OneQState
from pyquil.experiment._symmetrization import SymmetrizationLevel
from pyquil.gates import RESET
from pyquil.paulis import PauliTerm, is_identity
from pyquil.quil import Program
from pyquil.quilbase import Reset, ResetQubit

log = logging.getLogger(__name__)


def _abbrev_program(program: Program, max_len: int = 10) -> str:
    """Create an abbreviated string representation of a Program.

    This will join all instructions onto a single line joined by '; '. If the number of
    instructions exceeds ``max_len``, some will be excluded from the string representation.
    """
    program_lines = program.out().splitlines()
    if max_len is not None and len(program_lines) > max_len:
        first_n = max_len // 2
        last_n = max_len - first_n
        excluded = len(program_lines) - max_len
        program_lines = program_lines[:first_n] + [f"... {excluded} instrs not shown ..."] + program_lines[-last_n:]

    return "   " + "\n   ".join(program_lines)


def _remove_reset_from_program(program: Program) -> Program:
    """Trim the RESET from a program because in measure_observables it is re-added.

    :param program: Program to remove RESET(s) from.
    :return: Trimmed Program.
    """
    p = program.copy_everything_except_instructions()

    for inst in program:
        if not isinstance(inst, (Reset, ResetQubit)):
            p.inst(inst)

    return p


class Experiment:
    """A tomography-like experiment.

    Many near-term quantum algorithms involve:

     - some limited state preparation
     - enacting a quantum process (like in tomography) or preparing a variational ansatz state
       (like in VQE)
     - measuring observables of the state.

    Where we typically use a large number of (state_prep, measure) pairs but keep the ansatz
    program consistent. This class stores the ansatz program as a :py:class:`~pyquil.Program`
    and maintains a list of :py:class:`ExperimentSetting` objects which each represent a
    (state_prep, measure) pair.

    Settings diagonalized by a shared tensor product basis (TPB) can (optionally) be estimated
    simultaneously. Therefore, this class is backed by a list of list of ExperimentSettings.
    Settings sharing an inner list will be estimated simultaneously. If you don't want this,
    provide a list of length-1-lists. As a convenience, if you pass a 1D list to the constructor
    will expand it to a list of length-1-lists.

    This class will not group settings for you. Please see :py:func:`group_experiments` for
    a function that will automatically process a Experiment to group Experiments sharing
    a TPB.

    :ivar settings: The collection of ExperimentSetting objects that define this experiment.
    :ivar program: The main program body of this experiment. Also determines the ``shots``
        and ``reset`` instance variables. The ``shots`` instance variable is the number of
        shots to take per ExperimentSetting. The ``reset`` instance variable is whether to
        actively reset qubits instead of waiting several times the coherence length for qubits
        to decay to ``|0>`` naturally. Setting this to True is much faster but there is a ~1%
        error per qubit in the reset operation. Thermal noise from "traditional" reset is not
        routinely characterized but is of the same order.
    :ivar symmetrization: the level of readout symmetrization to perform for the estimation
        and optional calibration of each observable. The following integer levels, encapsulated in
        the ``SymmetrizationLevel`` integer enum, are currently supported:

        * -1 -- exhaustive symmetrization uses every possible combination of flips
        * 0 -- no symmetrization
        * 1 -- symmetrization using an orthogonal array (OA) with strength 1
        * 2 -- symmetrization using an orthogonal array (OA) with strength 2
        * 3 -- symmetrization using an orthogonal array (OA) with strength 3

        Note that (default) exhaustive symmetrization requires a number of QPU calls exponential in
        the number of qubits in the union of the support of the observables in any group of settings
        in ``tomo_experiment``; the number of shots may need to be increased to accommodate this.
        see :py:func:`run_symmetrized_readout` in api._quantum_computer for more information.
    """

    def __init__(
        self,
        settings: Union[list[ExperimentSetting], list[list[ExperimentSetting]]],
        program: Program,
        *,
        symmetrization: int = SymmetrizationLevel.EXHAUSTIVE,
        calibration: int = CalibrationMethod.PLUS_EIGENSTATE,
    ):
        if len(settings) == 0:
            s: list[list[ExperimentSetting]] = []
        else:
            if isinstance(settings[0], ExperimentSetting):
                # convenience wrapping in lists of length 1
                s = [[expt] for expt in cast(list[ExperimentSetting], settings)]
            else:
                s = cast(list[list[ExperimentSetting]], settings)

        self._settings = s
        self.program = program
        self.symmetrization = SymmetrizationLevel(symmetrization)
        if self.symmetrization != SymmetrizationLevel.EXHAUSTIVE:
            if isinstance(calibration, int) and calibration != 0:
                warnings.warn(
                    "Calibration is only supported for exhaustive symmetrization, "
                    "thus setting self.calibration = 0 (CalibrationMethod.NONE).",
                    stacklevel=2,
                )
            self.calibration = CalibrationMethod.NONE
        else:
            self.calibration = CalibrationMethod(calibration)
        self.shots = self.program.num_shots

        if "RESET" in self.program.out():
            self.reset = True
            self.program = _remove_reset_from_program(self.program)
        else:
            self.reset = False

    def __len__(self) -> int:
        return len(self._settings)

    def __getitem__(self, item: int) -> list[ExperimentSetting]:
        return self._settings[item]

    def __setitem__(self, key: int, value: list[ExperimentSetting]) -> None:
        self._settings[key] = value

    def __delitem__(self, key: int) -> None:
        self._settings.__delitem__(key)

    def __iter__(self) -> Generator[list[ExperimentSetting], None, None]:
        yield from self._settings

    def __reversed__(self) -> Generator[list[ExperimentSetting], None, None]:
        yield from reversed(self._settings)

    def __contains__(self, item: list[ExperimentSetting]) -> bool:
        return item in self._settings

    def append(self, expts: Union[ExperimentSetting, list[ExperimentSetting]]) -> None:
        if not isinstance(expts, list):
            expts = [expts]
        self._settings.append(expts)

    def count(self, expt: list[ExperimentSetting]) -> int:
        return self._settings.count(expt)

    def index(self, expt: list[ExperimentSetting], start: int = 0, stop: int = 0) -> int:
        return self._settings.index(expt, start, stop)

    def extend(self, expts: list[list[ExperimentSetting]]) -> None:
        self._settings.extend(expts)

    def insert(self, index: int, expt: list[ExperimentSetting]) -> None:
        self._settings.insert(index, expt)

    def pop(self, index: int = 0) -> list[ExperimentSetting]:
        return self._settings.pop(index)

    def remove(self, expt: list[ExperimentSetting]) -> None:
        self._settings.remove(expt)

    def reverse(self) -> None:
        self._settings.reverse()

    def sort(self, key: Optional[Callable[[list[ExperimentSetting]], Any]] = None, reverse: bool = False) -> None:
        return self._settings.sort(key=key, reverse=reverse)

    def setting_strings(self) -> Generator[str, None, None]:
        yield from (
            "{i}: {st_str}".format(i=i, st_str=", ".join(str(setting) for setting in settings))
            for i, settings in enumerate(self._settings)
        )

    def settings_string(self, abbrev_after: Optional[int] = None) -> str:
        setting_strs = list(self.setting_strings())
        if abbrev_after is not None and len(setting_strs) > abbrev_after:
            first_n = abbrev_after // 2
            last_n = abbrev_after - first_n
            excluded = len(setting_strs) - abbrev_after
            setting_strs = setting_strs[:first_n] + [f"... {excluded} settings not shown ..."] + setting_strs[-last_n:]
        return "   " + "\n   ".join(setting_strs)

    def __repr__(self) -> str:
        string = f"shots: {self.shots}\n"
        if self.reset:
            string += "active reset: enabled\n"
        else:
            string += "active reset: disabled\n"
        string += f"symmetrization: {self.symmetrization} ({self.symmetrization.name.lower()})\n"
        string += f"calibration: {self.calibration} ({self.calibration.name.lower()})\n"
        string += f"program:\n{_abbrev_program(self.program)}\n"
        string += f"settings:\n{self.settings_string(abbrev_after=20)}"
        return string

    def serializable(self) -> dict[str, Any]:
        return {
            "type": "Experiment",
            "settings": self._settings,
            "program": self.program.out(),
            "symmetrization": self.symmetrization,
            "shots": self.shots,
            "reset": self.reset,
        }

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Experiment):
            return False
        return self.serializable() == other.serializable()

    def get_meas_qubits(self) -> list[int]:
        """Return the sorted list of qubits involved in all out_operators of this Experiment object's settings."""
        meas_qubits: set[int] = set()
        for settings in self:
            if len(settings) != 1:
                raise ValueError("Can't get the measured qubits of grouped ExperimentSettings.")
            meas_qubits.update(cast(list[int], settings[0].out_operator.get_qubits()))
        return sorted(meas_qubits)

    def get_meas_registers(self, qubits: Optional[Sequence[int]] = None) -> list[int]:
        """Return the sorted list of memory registers corresponding to the list of qubits provided.

        If no qubits are provided, just returns the list of numbers from 0 to n-1 where n is the
        number of qubits resulting from the ``get_meas_qubits`` method.
        """
        meas_qubits = self.get_meas_qubits()

        if qubits is None:
            return list(range(len(meas_qubits)))

        meas_registers = []
        for q in qubits:
            meas_registers.append(meas_qubits.index(q))
        return sorted(meas_registers)

    def generate_experiment_program(self) -> Program:
        """Generate a parameterized program with the main body and additions for state prep, measurement, and symmetrization.

        State preparation and measurement are achieved via ZXZXZ-decomposed single-qubit gates,
        where the angles of each ``RZ`` rotation are declared parameters that can be assigned at
        runtime. Symmetrization is achieved by putting an ``RX`` gate (also parameterized by a
        declared value) before each ``MEASURE`` operation. In addition, a ``RESET`` operation
        is prepended to the ``Program`` if the experiment has active qubit reset enabled. Finally,
        each qubit specified in the settings is measured, and the number of shots is added.

        :return: Parameterized ``Program`` that is capable of collecting statistics for every
            ``ExperimentSetting`` in this ``Experiment``.
        """
        meas_qubits = self.get_meas_qubits()

        p = Program()

        if self.reset:
            if any(isinstance(instr, (Reset, ResetQubit)) for instr in self.program):
                raise ValueError("RESET already added to program")
            p += RESET()

        for settings in self:
            if len(settings) != 1:
                raise ValueError("Can't generate a program for grouped ExperimentSettings.")
            if ("X" in str(settings[0].in_state)) or ("Y" in str(settings[0].in_state)):
                if "DECLARE preparation_alpha" in self.program.out():
                    raise ValueError('Memory "preparation_alpha" has been declared already.')
                if "DECLARE preparation_beta" in self.program.out():
                    raise ValueError('Memory "preparation_beta" has been declared already.')
                if "DECLARE preparation_gamma" in self.program.out():
                    raise ValueError('Memory "preparation_gamma" has been declared already.')
                p += parameterized_single_qubit_state_preparation(meas_qubits)
                break

        p += self.program

        for settings in self:
            if len(settings) != 1:
                raise ValueError("Can't generate a program for grouped ExperimentSettings.")
            if ("X" in str(settings[0].out_operator)) or ("Y" in str(settings[0].out_operator)):
                if "DECLARE measurement_alpha" in self.program.out():
                    raise ValueError('Memory "measurement_alpha" has been declared already.')
                if "DECLARE measurement_beta" in self.program.out():
                    raise ValueError('Memory "measurement_beta" has been declared already.')
                if "DECLARE measurement_gamma" in self.program.out():
                    raise ValueError('Memory "measurement_gamma" has been declared already.')
                p += parameterized_single_qubit_measurement_basis(meas_qubits)
                break

        if self.symmetrization != 0:
            if "DECLARE symmetrization" in self.program.out():
                raise ValueError('Memory "symmetrization" has been declared already.')
            p += parameterized_readout_symmetrization(meas_qubits)

        if "DECLARE ro" in self.program.out():
            raise ValueError('Memory "ro" has already been declared for this program.')
        p += measure_qubits(meas_qubits)

        p.wrap_in_numshots_loop(self.shots)

        return p

    def build_setting_memory_map(self, setting: ExperimentSetting) -> dict[str, list[float]]:
        """Build the memory map for state prep and measurement specs in the ExperimentSetting, considering all qubits.

        The memory map is built corresponding to the state preparation and measurement specifications encoded in the
        provided ExperimentSetting, taking into account the full set of qubits present in the Experiment object.

        :return: Memory map for state prep and measurement.
        """
        meas_qubits = self.get_meas_qubits()

        in_pt = PauliTerm.from_list([(op, meas_qubits.index(cast(int, q))) for q, op in setting._in_operator()])
        out_pt = PauliTerm.from_list([(op, meas_qubits.index(cast(int, q))) for q, op in setting.out_operator])

        preparation_map = pauli_term_to_preparation_memory_map(in_pt)
        measurement_map = pauli_term_to_measurement_memory_map(out_pt)

        return {**preparation_map, **measurement_map}

    def build_symmetrization_memory_maps(
        self, qubits: Sequence[int], label: str = "symmetrization"
    ) -> list[dict[str, list[float]]]:
        """Build a list of memory maps for readout symmetrization in a program using parametric compilation.

        For example, if we have the following program:

            RX(symmetrization[0]) 0
            RX(symmetrization[1]) 1
            MEASURE 0 ro[0]
            MEASURE 1 ro[1]

        We can perform exhaustive readout symmetrization on our two qubits by providing the four
        following memory maps, and then appropriately flipping the resultant bitstrings:

            {'symmetrization': [0.0, 0.0]} -> XOR results with [0,0]
            {'symmetrization': [0.0, pi]}  -> XOR results with [0,1]
            {'symmetrization': [pi, 0.0]}  -> XOR results with [1,0]
            {'symmetrization': [pi, pi]}   -> XOR results with [1,1]

        :param qubits: List of qubits to symmetrize readout for.
        :param label: Name of the declared memory region. Defaults to "symmetrization".
        :return: List of memory maps that performs the desired level of symmetrization.
        """
        num_meas_registers = len(self.get_meas_qubits())
        symm_registers = self.get_meas_registers(qubits)

        if self.symmetrization == SymmetrizationLevel.NONE:
            return [{}]

        # TODO: add support for orthogonal arrays
        if self.symmetrization != SymmetrizationLevel.EXHAUSTIVE:
            raise ValueError("We only support exhaustive symmetrization for now.")

        import itertools

        import numpy as np

        assignments = itertools.product(np.array([0, np.pi]), repeat=len(symm_registers))
        memory_maps = []
        for a in assignments:
            zeros = np.zeros(num_meas_registers)
            for idx, r in enumerate(symm_registers):
                zeros[r] = a[idx]
            memory_maps.append({f"{label}": list(zeros)})
        return memory_maps

    def generate_calibration_experiment(self) -> "Experiment":
        """Generate an Experiment for calibrating multi-qubit observables to mitigate readout errors.

        Generate another Experiment object to calibrate the various multi-qubit observables involved in this Experiment.
        This is done by preparing the plus-one (minus-one) eigenstate of each out_operator and measuring the resulting
        expectation value of the same out_operator. Ideally, this should always yield +1 (-1), but the presence of
        symmetric readout error scales the results by a constant factor. Determining this scale factor is known as
        readout calibration, allowing for error mitigation in subsequent measurements by dividing by the scale factor.

        :return: A new ``Experiment`` that can calibrate the readout error of all the observables involved in
            this experiment.
        """
        if self.calibration != CalibrationMethod.PLUS_EIGENSTATE:
            raise ValueError('We currently only support the "plus eigenstate" calibration method.')

        calibration_settings = []
        for settings in self:
            if len(settings) != 1:
                raise ValueError("Can't generate calibration experiment with grouped ExperimentSettings.")
            calibration_settings.append(
                ExperimentSetting(
                    in_state=_pauli_to_product_state(settings[0].out_operator),
                    out_operator=settings[0].out_operator,
                    additional_expectations=settings[0].additional_expectations,
                )
            )

        calibration_program = Program()
        if self.reset:
            calibration_program += RESET()
        calibration_program.wrap_in_numshots_loop(self.shots)

        if self.symmetrization != SymmetrizationLevel.EXHAUSTIVE:
            raise ValueError("We currently only support calibration for exhaustive symmetrization")

        return Experiment(
            settings=calibration_settings,
            program=calibration_program,
            symmetrization=SymmetrizationLevel.EXHAUSTIVE,
            calibration=CalibrationMethod.NONE,
        )


def _pauli_to_product_state(in_state: PauliTerm) -> TensorProductState:
    """Convert a Pauli term to a TensorProductState."""
    if is_identity(in_state):
        return TensorProductState()
    else:
        return TensorProductState(
            [
                _OneQState(label=pauli_label, index=0, qubit=cast(int, qubit))
                for qubit, pauli_label in in_state._ops.items()
            ]
        )


class OperatorEncoder(JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, ExperimentSetting):
            return o.serializable()
        if isinstance(o, Experiment):
            return o.serializable()
        if isinstance(o, ExperimentResult):
            return o.serializable()
        return o


def to_json(fn: str, obj: Any) -> str:
    """Save pyquil.experiment objects as a JSON file.

    See :py:func:`read_json`.
    """
    # Specify UTF-8 to guard against systems that default to an ASCII locale.
    with open(fn, "w", encoding="utf-8") as f:
        json.dump(obj, f, cls=OperatorEncoder, indent=2, ensure_ascii=False)
    return fn


def _operator_object_hook(obj: Mapping[str, Any]) -> Union[Mapping[str, Any], Experiment]:
    if "type" in obj and obj["type"] in ["Experiment", "TomographyExperiment"]:
        # I bet this doesn't work for grouped experiment settings
        settings = [[ExperimentSetting.from_str(s) for s in stt] for stt in obj["settings"]]
        p = Program(obj["program"])
        p.wrap_in_numshots_loop(obj["shots"])
        ex = Experiment(settings=settings, program=p, symmetrization=obj["symmetrization"])
        ex.reset = obj["reset"]
        return ex
    return obj


def read_json(fn: str) -> Any:
    """Read pyquil.experiment objects from a JSON file.

    See :py:func:`to_json`.
    """
    # Specify UTF-8 to guard against systems that default to an ASCII locale.
    with open(fn, encoding="utf-8") as f:
        return json.load(f, object_hook=_operator_object_hook)



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""Module for working with Pauli algebras."""

import copy
import re
import warnings
from collections import OrderedDict
from collections.abc import Hashable, Iterable, Iterator, Sequence
from functools import reduce
from itertools import product
from numbers import Complex, Number
from typing import (
    Callable,
    Optional,
    Union,
    cast,
)

import numpy as np
import quil.instructions as quil_rs
from numpy.typing import NDArray
from scipy.linalg import expm

from pyquil.quilatom import (
    Expression,
    ExpressionDesignator,
    FormalArgument,
    QubitDesignator,
    QubitPlaceholder,
    _convert_to_py_expression,
    _convert_to_py_qubits,
    _convert_to_rs_expression,
)

from .gates import CNOT, PHASE, QUANTUM_GATES, RX, RZ, H, X
from .quil import Program

PauliTargetDesignator = QubitDesignator
PauliDesignator = Union["PauliTerm", "PauliSum"]

PAULI_OPS = ["X", "Y", "Z", "I"]
PAULI_PROD = {
    "ZZ": "I",
    "YY": "I",
    "XX": "I",
    "II": "I",
    "XY": "Z",
    "XZ": "Y",
    "YX": "Z",
    "YZ": "X",
    "ZX": "Y",
    "ZY": "X",
    "IX": "X",
    "IY": "Y",
    "IZ": "Z",
    "ZI": "Z",
    "YI": "Y",
    "XI": "X",
    "X": "X",
    "Y": "Y",
    "Z": "Z",
    "I": "I",
}
PAULI_COEFF = {
    "ZZ": 1.0,
    "YY": 1.0,
    "XX": 1.0,
    "II": 1.0,
    "XY": 1.0j,
    "XZ": -1.0j,
    "YX": -1.0j,
    "YZ": 1.0j,
    "ZX": 1.0j,
    "ZY": -1.0j,
    "IX": 1.0,
    "IY": 1.0,
    "IZ": 1.0,
    "ZI": 1.0,
    "YI": 1.0,
    "XI": 1.0,
    "X": 1.0,
    "Y": 1.0,
    "Z": 1.0,
    "I": 1.0,
}


class UnequalLengthWarning(Warning):
    """Warning issued when multiplying PauliTerms of different lengths."""

    def __init__(self, *args: object, **kwargs: object):
        """Initialize the warning."""
        super().__init__(*args, **kwargs)


integer_types = (int, np.int64, np.int32, np.int16, np.int8)
"""Explicitly include numpy integer dtypes (for python 3)."""

HASH_PRECISION = 1e6
"""The precision used when hashing terms to check equality. The simplify() method
uses np.isclose() for coefficient comparisons to 0 which has its own default precision. We
can't use np.isclose() for hashing terms though.
"""


def _valid_qubit(index: Optional[Union[PauliTargetDesignator, QubitPlaceholder]]) -> bool:
    return (
        (isinstance(index, integer_types) and index >= 0)
        or isinstance(index, QubitPlaceholder)
        or isinstance(index, FormalArgument)
    )


class PauliTerm:
    """A term is a product of Pauli operators operating on different qubits."""

    def __init__(
        self,
        op: str,
        index: Optional[PauliTargetDesignator],
        coefficient: ExpressionDesignator = 1.0,
    ):
        """Create a new Pauli Term with a Pauli operator at a particular index and a leading coefficient.

        :param op: The Pauli operator as a string "X", "Y", "Z", or "I"
        :param index: The qubit index that that operator is applied to.
        :param coefficient: The coefficient multiplying the operator, e.g. 1.5 * Z_1
        """
        if op not in PAULI_OPS:
            raise ValueError(f"{op} is not a valid Pauli operator")

        self._ops: dict[PauliTargetDesignator, str] = OrderedDict()
        if op != "I":
            if not _valid_qubit(index):
                raise ValueError(f"{index} is not a valid qubit")
            if index is None:
                raise ValueError("PauliTerm index cannot be None")
            self._ops[index] = op

        if isinstance(coefficient, Number):
            self.coefficient: Union[complex, Expression] = complex(coefficient)
        else:
            self.coefficient = coefficient

    @classmethod
    def _from_rs_pauli_term(cls, term: quil_rs.PauliTerm) -> "PauliTerm":
        term_list = [(str(gate), FormalArgument(arg)) for (gate, arg) in term.arguments]
        coefficient = _convert_to_py_expression(term.expression.into_simplified())

        return cls.from_list(term_list, coefficient)

    def _to_rs_pauli_term(self) -> quil_rs.PauliTerm:
        arguments = [(quil_rs.PauliGate.parse(gate), str(arg)) for arg, gate in self._ops.items()]
        return quil_rs.PauliTerm(arguments, _convert_to_rs_expression(self.coefficient))

    def id(self, sort_ops: bool = True) -> str:
        """Return an identifier string for the PauliTerm (ignoring the coefficient).

        Don't use this to compare terms. This function will not work with qubits that
        aren't sortable.

        :param sort_ops: Whether to sort operations by qubit. This is True by default for
            backwards compatibility but will change in a future version. Callers should never rely
            on comparing id's for testing equality. See ``operations_as_set`` instead.
        :return: A string representation of this term's operations.
        """
        if len(self._ops) == 0 and not sort_ops:
            # This is nefariously backwards-compatibility breaking. There's potentially
            # lots of code floating around that says is_identity = term.id() == ''
            # Please use `is_identity(term)`!
            # Therefore, we only return 'I' when sort_ops is set to False, which is the newer
            # way of calling this function and implies the user knows what they're doing.
            return "I"

        if sort_ops and len(self._ops) > 1:
            warnings.warn(
                "`PauliTerm.id()` will not work on PauliTerms where the qubits are not "
                "sortable and should be avoided in favor of `operations_as_set`.",
                FutureWarning,
                stacklevel=2,
            )
            return "".join(f"{self._ops[q]}{q}" for q in sorted(self._ops.keys()))  # type: ignore
        else:
            return "".join(f"{p}{q}" for q, p in self._ops.items())

    def operations_as_set(self) -> frozenset[tuple[PauliTargetDesignator, str]]:
        """Return a frozenset of operations in this term.

        Use this in place of :py:func:`id` if the order of operations in the term does not
        matter.

        :return: frozenset of (qubit, op_str) representing Pauli operations
        """
        return frozenset(self._ops.items())

    def __eq__(self, other: object) -> bool:
        """Return True if two PauliTerms are equal."""
        if not isinstance(other, (PauliTerm, PauliSum)):
            raise TypeError(f"Can't compare PauliTerm with object of type {type(other)}.")
        elif isinstance(other, PauliSum):
            return other == self
        else:
            return self.operations_as_set() == other.operations_as_set() and np.allclose(
                self.coefficient, other.coefficient
            )

    def __hash__(self) -> int:
        """Return a hash of the PauliTerm."""
        if not isinstance(self.coefficient, Complex):
            raise ValueError("PauliTerm cannot be hashed if the coefficient is not a complex number.")
        return hash(
            (
                round(self.coefficient.real * HASH_PRECISION),
                round(self.coefficient.imag * HASH_PRECISION),
                self.operations_as_set(),
            )
        )

    def __len__(self) -> int:
        """Length of the PauliTerm is the number of Pauli operators in the term.

        A term that consists of only a scalar has a length of zero.
        """
        return len(self._ops)

    def copy(self) -> "PauliTerm":
        """Create a new PauliTerm, with a completely new dictionary of operators."""
        new_term = PauliTerm("I", 0, 1.0)  # create new object
        # manually copy all attributes over
        for key in self.__dict__.keys():
            val = self.__dict__[key]
            if isinstance(val, (dict, list, set)):  # mutable types
                new_term.__dict__[key] = copy.copy(val)
            else:  # immutable types
                new_term.__dict__[key] = val

        return new_term

    @property
    def program(self) -> Program:
        """Create Program from the PauliTerm."""
        return Program([QUANTUM_GATES[gate](q) for q, gate in self])

    def get_qubits(self) -> list[PauliTargetDesignator]:
        """Get all the qubits that this PauliTerm operates on."""
        return list(self._ops.keys())

    def __getitem__(self, i: PauliTargetDesignator) -> str:
        """Get the Pauli operator for a particular target."""
        return self._ops.get(i, "I")

    def __iter__(self) -> Iterator[tuple[PauliTargetDesignator, str]]:
        """Iterate over the quibit indices and Pauli operators in the term."""
        for i in self.get_qubits():
            yield i, self[i]

    def _multiply_factor(self, factor: str, index: PauliTargetDesignator) -> "PauliTerm":
        new_term = PauliTerm("I", 0)
        new_coeff = self.coefficient
        new_ops = self._ops.copy()

        ops = self[index] + factor
        new_op = PAULI_PROD[ops]
        if new_op != "I":
            new_ops[index] = new_op
        else:
            del new_ops[index]
        new_coeff *= PAULI_COEFF[ops]

        new_term._ops = new_ops
        new_term.coefficient = new_coeff

        return new_term

    def __mul__(self, term: Union[PauliDesignator, ExpressionDesignator]) -> PauliDesignator:
        """Multiply this Pauli Term with another PauliTerm, PauliSum, or number according to the Pauli algebra rules.

        :param term: (PauliTerm or PauliSum or Number) A term to multiply by.
        :returns: The product of this PauliTerm and term.
        """
        if isinstance(term, PauliSum):
            return (PauliSum([self]) * term).simplify()
        elif isinstance(term, PauliTerm):
            new_term = PauliTerm("I", 0, 1.0)
            new_term._ops = self._ops.copy()
            new_coeff = self.coefficient * term.coefficient
            for index, op in term:
                new_term = new_term._multiply_factor(op, index)

            return term_with_coeff(new_term, new_term.coefficient * new_coeff)
        return term_with_coeff(self, self.coefficient * term)

    def __rmul__(self, other: ExpressionDesignator) -> "PauliTerm":
        """Multiplies this PauliTerm with another object, probably a number.

        :param other: A number or PauliTerm to multiply by
        :returns: A new PauliTerm
        """
        p = self * other
        if not isinstance(p, PauliTerm):
            raise ValueError(f"PauliTerm multiplication should return a PauliTerm, got {type(other)}")
        return p

    def __pow__(self, power: int) -> "PauliTerm":
        """Raise the PauliTerm to power.

        :param power: The power to raise this PauliTerm to.
        :return: The power-fold product of power.
        """
        if not isinstance(power, int) or power < 0:
            raise ValueError("The power must be a non-negative integer.")

        if len(self.get_qubits()) == 0:
            # There weren't any nontrivial operators
            return term_with_coeff(self, 1)

        result = ID()
        for _ in range(power):
            result = cast(PauliTerm, result * self)
        return result

    def __add__(self, other: Union[PauliDesignator, ExpressionDesignator]) -> "PauliSum":
        """Add this PauliTerm with another one.

        :param other: A PauliTerm object, a PauliSum object, or a Number
        :returns: A PauliSum object representing the sum of this PauliTerm and other
        """
        if isinstance(other, PauliSum):
            return other + self
        elif isinstance(other, PauliTerm):
            new_sum = PauliSum([self, other])
            return new_sum.simplify()
        else:  # is a Number
            return self + PauliTerm("I", 0, other)

    def __radd__(self, other: ExpressionDesignator) -> "PauliSum":
        """Add this PauliTerm with a Number.

        :param other: A Number
        :returns: A new PauliSum
        """
        return PauliTerm("I", 0, other) + self

    def __sub__(self, other: Union["PauliTerm", ExpressionDesignator]) -> "PauliSum":
        """Subtracts a PauliTerm from this one.

        :param other: A PauliTerm object, a number, or an Expression
        :returns: A PauliSum object representing the difference of this PauliTerm and term
        """
        return self + -1.0 * other

    def __rsub__(self, other: Union["PauliTerm", ExpressionDesignator]) -> "PauliSum":
        """Subtracts this PauliTerm from a Number or PauliTerm.

        :param other: A PauliTerm object or a Number
        :returns: A PauliSum object representing the difference of this PauliTerm and term
        """
        return other + -1.0 * self

    def __repr__(self) -> str:
        """Get a string representation of the PauliTerm."""
        term_strs = []
        for index in self._ops.keys():
            term_strs.append(f"{self[index]}{index}")

        if len(term_strs) == 0:
            term_strs.append("I")
        out = f"{self.coefficient}*{'*'.join(term_strs)}"
        return out

    def compact_str(self) -> str:
        """Return string representation of the Pauli term that is more compact than ``str(term)``.

        >>> term = 2.0 * sX(1) * sZ(2)
        >>> str(term)
        '(2+0j)*X1*Z2'
        >>> term.compact_str()
        '(2+0j)*X1Z2'
        """
        return f"{self.coefficient}*{self.id(sort_ops=False)}"

    @classmethod
    def from_list(
        cls, terms_list: Sequence[tuple[str, PauliTargetDesignator]], coefficient: ExpressionDesignator = 1.0
    ) -> "PauliTerm":
        """Allocate a PauliTerm from a list of operators and indices.

        This is more efficient than multiplying together individual terms.

        :param list terms_list: A list of tuples, e.g. [("X", 0), ("Y", 1)]
        :return: PauliTerm
        """
        if not all([isinstance(op, tuple) for op in terms_list]):
            raise TypeError(
                "The type of terms_list should be a list of (name, index) " "tuples suitable for PauliTerm()."
            )

        pterm = PauliTerm("I", 0)
        if not all([op[0] in PAULI_OPS for op in terms_list]):
            raise ValueError("All operators must be in ['X', 'Y', 'Z', 'I']")

        indices = [op[1] for op in terms_list]
        if not all(_valid_qubit(index) for index in indices):
            raise ValueError("All qubit indices must be valid.")

        # this is because from_list doesn't call simplify in order to be more efficient.
        if len(set(indices)) != len(indices):
            raise ValueError(
                "Elements of PauliTerm that are allocated using from_list must "
                "be on disjoint qubits. Use PauliTerm multiplication to simplify "
                "terms instead."
            )

        for op, index in terms_list:
            if op != "I":
                pterm._ops[index] = op
        if isinstance(coefficient, Number):
            pterm.coefficient = complex(coefficient)
        else:
            pterm.coefficient = coefficient
        return pterm

    @classmethod
    def from_compact_str(cls, str_pauli_term: str) -> "PauliTerm":
        """Construct a PauliTerm from the result of str(pauli_term)."""
        # split into str_coef, str_op at first '*'' outside parenthesis
        try:
            str_coef, str_op = re.split(r"\*(?![^(]*\))", str_pauli_term, maxsplit=1)
        except ValueError as e:
            raise ValueError(
                "Could not separate the pauli string into "
                f"coefficient and operator. {str_pauli_term} does"
                " not match <coefficient>*<operator>"
            ) from e

        # parse the coefficient into either a float or complex
        str_coef = str_coef.replace(" ", "")
        try:
            coef: Union[float, complex] = float(str_coef)
        except ValueError:
            try:
                coef = complex(str_coef)
            except ValueError as e:
                raise ValueError(f"Could not parse the coefficient {str_coef}") from e

        op = sI() * coef
        if str_op == "I":
            if not isinstance(op, PauliTerm):
                raise ValueError(f"Expected operation to be PauliTerm, got {type(op)}.")
            return op

        # parse the operator
        str_op = re.sub(r"\*", "", str_op)
        if not re.match(r"^(([XYZ])(\d+))+$", str_op):
            raise ValueError(rf"Could not parse operator string {str_op}. It should match ^(([XYZ])(\d+))+$")

        for factor in re.finditer(r"([XYZ])(\d+)", str_op):
            op *= cls(factor.group(1), int(factor.group(2)))

        if not isinstance(op, PauliTerm):
            raise ValueError(f"Expected operation to be PauliTerm, got {type(op)}.")
        return op

    def pauli_string(self, qubits: Optional[Iterable[int]] = None) -> str:
        """Return a string representation of this PauliTerm without its coefficient and with implicit qubit indices.

        If an iterable of qubits is provided, each character in the resulting string represents
        a Pauli operator on the corresponding qubit.

        >>> p = PauliTerm("X", 0) * PauliTerm("Y", 1, 1.0j)
        >>> p.pauli_string()
        'XY'
        >>> p.pauli_string(qubits=[0])
        'X'
        >>> p.pauli_string(qubits=[0, 2])
        'XI'

        :param iterable of qubits: The iterable of qubits to represent, given as ints.
        :return: The string representation of this PauliTerm, sans coefficient
        """
        if qubits is None:
            return "".join(self._ops.values())
        return "".join(self[q] for q in qubits)


# For convenience, a shorthand for several operators.
def ID() -> PauliTerm:
    """Identity operator."""
    return PauliTerm("I", 0, 1)


def ZERO() -> PauliTerm:
    """Zero operator."""
    return PauliTerm("I", 0, 0)


def sI(q: Optional[int] = None) -> PauliTerm:
    """Return the identity operator, optionally on a particular qubit.

    This can be specified without a qubit.

    :param qubit_index: The optional index of a qubit.
    :returns: A PauliTerm object
    """
    return PauliTerm("I", q)


def sX(q: int) -> PauliTerm:
    """Return the sigma_X operator on a particular qubit.

    :param qubit_index: The index of the qubit
    :returns: A PauliTerm object
    """
    return PauliTerm("X", q)


def sY(q: int) -> PauliTerm:
    """Return the sigma_Y operator on a particular qubit.

    :param qubit_index: The index of the qubit
    :returns: A PauliTerm object
    """
    return PauliTerm("Y", q)


def sZ(q: int) -> PauliTerm:
    """Return the sigma_Z operator on a particular qubit.

    :param qubit_index: The index of the qubit
    :returns: A PauliTerm object
    """
    return PauliTerm("Z", q)


def term_with_coeff(term: PauliTerm, coeff: ExpressionDesignator) -> PauliTerm:
    """Change the coefficient of a PauliTerm.

    :param term: A PauliTerm object
    :param coeff: The coefficient to set on the PauliTerm
    :returns: A new PauliTerm that duplicates term but sets coeff
    """
    if not isinstance(coeff, Number):
        raise ValueError("coeff must be a Number")
    new_pauli = term.copy()
    # We cast to a complex number to ensure that internally the coefficients remain compatible.
    new_pauli.coefficient = complex(coeff)
    return new_pauli


class PauliSum:
    """A sum of one or more PauliTerms."""

    def __init__(self, terms: Sequence[PauliTerm]):
        """Initialize a PauliSum with a list of PauliTerms.

        :param Sequence terms: A Sequence of PauliTerms.
        """
        if not (isinstance(terms, Sequence) and all([isinstance(term, PauliTerm) for term in terms])):
            raise ValueError("PauliSum's are currently constructed from Sequences of PauliTerms.")
        self.terms: Sequence[PauliTerm]
        if len(terms) == 0:
            self.terms = [0.0 * ID()]
        else:
            self.terms = terms

    @classmethod
    def _from_rs_pauli_sum(cls, pauli_sum: quil_rs.PauliSum) -> "PauliSum":
        return cls([PauliTerm._from_rs_pauli_term(term) for term in pauli_sum.terms])

    def _to_rs_pauli_sum(self, arguments: Optional[list[PauliTargetDesignator]] = None) -> quil_rs.PauliSum:
        rs_arguments: list[str]
        if arguments is None:
            argument_set: dict[str, None] = {}
            for term_arguments in [term.get_qubits() for term in self.terms]:
                argument_set.update({str(arg): None for arg in term_arguments})
            rs_arguments = list(argument_set.keys())
        else:
            rs_arguments = [str(arg) for arg in arguments]
        terms = [term._to_rs_pauli_term() for term in self.terms]
        return quil_rs.PauliSum(rs_arguments, terms)

    def __eq__(self, other: object) -> bool:
        """Equality testing to see if two PauliSum's are equivalent.

        :param other: The PauliSum to compare this PauliSum with.
        :return: True if other is equivalent to this PauliSum, False otherwise.
        """
        if not isinstance(other, (PauliTerm, PauliSum)):
            raise TypeError(f"Can't compare PauliSum with object of type {type(other)}.")
        elif isinstance(other, PauliTerm):
            return self == PauliSum([other])
        elif len(self.terms) != len(other.terms):
            return False

        return set(self.terms) == set(other.terms)

    def __hash__(self) -> int:
        """Return a hash value for the sum."""
        return hash(frozenset(self.terms))

    def __repr__(self) -> str:
        """Return a string representation for the sum."""
        return " + ".join([str(term) for term in self.terms])

    def __len__(self) -> int:
        """Return the number of PauliTerms in the sum."""
        return len(self.terms)

    def __getitem__(self, item: int) -> PauliTerm:
        """Get the PauliTerm at the given position the sum.

        :param item: The index of the term in the sum to return
        :return: The PauliTerm at the index-th position in the PauliSum
        """
        return self.terms[item]

    def __iter__(self) -> Iterator[PauliTerm]:
        """Iterate over the PauliTerms in the sum."""
        return self.terms.__iter__()

    def __mul__(self, other: Union[PauliDesignator, ExpressionDesignator]) -> "PauliSum":
        """Multiply and simplify this PauliSum with another PauliSum, PauliTerm or Number object.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the multiplication.
        """
        if not isinstance(other, (Expression, Number, PauliTerm, PauliSum)):
            raise ValueError("Cannot multiply PauliSum by term that is not a Number, PauliTerm, or PauliSum")

        other_terms: list[Union[PauliTerm, ExpressionDesignator]] = []
        if isinstance(other, PauliSum):
            other_terms += other.terms
        else:
            other_terms += [other]

        new_terms = [lterm * rterm for lterm, rterm in product(self.terms, other_terms)]
        new_sum = PauliSum(cast(list[PauliTerm], new_terms))
        return new_sum.simplify()

    def __rmul__(self, other: ExpressionDesignator) -> "PauliSum":
        """Multiply and simplify this PauliSum with another PauliSum, PauliTerm or Number object.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the multiplication.
        """
        if not isinstance(other, Number):
            raise TypeError(f"Expected a Number object, got {type(other)}")
        new_terms = [term.copy() for term in self.terms]
        for term in new_terms:
            term.coefficient *= other
        return PauliSum(new_terms).simplify()

    def __pow__(self, power: int) -> "PauliSum":
        """Raise this PauliSum to power.

        :param power: The power to raise this PauliSum to.
        :return: The power-th power of this PauliSum.
        """
        if not isinstance(power, int) or power < 0:
            raise ValueError("The power must be a non-negative integer.")
        result = PauliSum([ID()])

        if not self.get_qubits():
            # There aren't any nontrivial operators
            terms = [term_with_coeff(term, 1) for term in self.terms]
            for term in terms:
                result *= term
        else:
            for term in self.terms:
                for qubit_id in term.get_qubits():
                    result *= PauliTerm("I", qubit_id)

        for _ in range(power):
            result *= self
        return result

    def __add__(self, other: Union[PauliDesignator, ExpressionDesignator]) -> "PauliSum":
        """Add and simplify this PauliSum with another PauliSum, PauliTerm or Number objects.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the addition.
        """
        if isinstance(other, PauliTerm):
            other_sum = PauliSum([other])
        elif isinstance(other, (Expression, Number, complex)):
            other_sum = PauliSum([other * ID()])
        elif isinstance(other, PauliSum):
            other_sum = other
        else:
            raise TypeError(f"{type(other)} is not a valid PauliDesignator or ExpressionDesignator")
        new_terms = [term.copy() for term in self.terms]
        new_terms.extend(other_sum.terms)
        new_sum = PauliSum(new_terms)
        return new_sum.simplify()

    def __radd__(self, other: ExpressionDesignator) -> "PauliSum":
        """Add and simplify this PauliSum with another PauliSum, PauliTerm or Number object.

        :param other: A Number
        :return: A new PauliSum object given by the addition.
        """
        if not isinstance(other, Number):
            raise TypeError(f"Expected a Number object, got {type(other)}")
        return self + other

    def __sub__(self, other: Union[PauliDesignator, ExpressionDesignator]) -> "PauliSum":
        """Subtract and simplify this PauliSum with another PauliSum, PauliTerm or Number object.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the subtraction.
        """
        return self + -1.0 * other

    def __rsub__(self, other: Union[PauliDesignator, ExpressionDesignator]) -> "PauliSum":
        """Subtract and simplify this PauliSum with another PauliSum, PauliTerm or Number object.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the subtraction.
        """
        return other + -1.0 * self

    def get_qubits(self) -> list[PauliTargetDesignator]:
        """Get the qubits that this PauliSum operates on.

        :returns: A list of all the qubits in the sum of terms.
        """
        all_qubits: set[PauliTargetDesignator] = set()
        for term in self.terms:
            all_qubits.update(term.get_qubits())
        return _convert_to_py_qubits(set(all_qubits))

    def simplify(self) -> "PauliSum":
        """Simplify the sum of Pauli operators according to Pauli algebra rules."""
        return simplify_pauli_sum(self)

    def get_programs(self) -> tuple[list[Program], np.ndarray]:
        """Get a Pyquil Program corresponding to each term in the PauliSum and a coefficient for each program.

        :return: (programs, coefficients)
        """
        programs = [term.program for term in self.terms]
        coefficients = np.array([term.coefficient for term in self.terms])
        return programs, coefficients

    def compact_str(self) -> str:
        """Return a string representation of the PauliSum that is more compact than ``str(pauli_sum)``.

        >>> pauli_sum = 2.0 * sX(1) * sZ(2) + 1.5 * sY(2)
        >>> str(pauli_sum)
        '(2+0j)*X1*Z2 + (1.5+0j)*Y2'
        >>> pauli_sum.compact_str()
        '(2+0j)*X1Z2+(1.5+0j)*Y2'
        """
        return "+".join([term.compact_str() for term in self.terms])

    @classmethod
    def from_compact_str(cls, str_pauli_sum: str) -> "PauliSum":
        """Construct a PauliSum from the result of str(pauli_sum)."""
        # split str_pauli_sum only at "+" outside of parenthesis to allow
        # e.g. "0.5*X0 + (0.5+0j)*Z2"
        str_terms = re.split(r"\+(?![^(]*\))", str_pauli_sum)
        str_terms = [s.strip() for s in str_terms]
        terms = [PauliTerm.from_compact_str(term) for term in str_terms]
        return cls(terms).simplify()


def simplify_pauli_sum(pauli_sum: PauliSum) -> PauliSum:
    """Simplify the sum of Pauli operators according to Pauli algebra rules.

    Warning: The simplified expression may re-order pauli operations, and may
    impact the observed performance when running on the QPU.
    """
    # You might want to use a defaultdict(list) here, but don't because
    # we want to do our best to preserve the order of terms.
    like_terms: dict[Hashable, list[PauliTerm]] = OrderedDict()
    for term in pauli_sum.terms:
        key = term.operations_as_set()
        if key in like_terms:
            like_terms[key].append(term)
        else:
            like_terms[key] = [term]

    terms = []
    for term_list in like_terms.values():
        first_term = term_list[0]
        if len(term_list) == 1 and not np.isclose(first_term.coefficient, 0.0):
            terms.append(first_term)
        else:
            coeff = sum(t.coefficient for t in term_list)
            if not np.isclose(coeff, 0.0):
                terms.append(term_with_coeff(term_list[0], coeff))
    return PauliSum(terms)


def check_commutation(pauli_list: Sequence[PauliTerm], pauli_two: PauliTerm) -> bool:
    """Check if commuting a PauliTerm commutes with a list of other terms by natural calculation.

    Uses the result in Section 3 of arXiv:1405.5749v2, modified slightly here to check for the
    number of anti-coincidences (which must always be even for commuting PauliTerms)
    instead of the no. of coincidences, as in the paper.

    :param pauli_list: A list of PauliTerm objects
    :param pauli_two_term: A PauliTerm object
    :returns: True if pauli_two object commutes with pauli_list, False otherwise
    """

    def coincident_parity(p1: PauliTerm, p2: PauliTerm) -> bool:
        non_similar = 0
        p1_indices = set(p1._ops.keys())
        p2_indices = set(p2._ops.keys())
        for idx in p1_indices.intersection(p2_indices):
            if p1[idx] != p2[idx]:
                non_similar += 1
        return non_similar % 2 == 0

    for term in pauli_list:
        if not coincident_parity(term, pauli_two):
            return False
    return True


def commuting_sets(pauli_terms: PauliSum) -> list[list[PauliTerm]]:
    """Gather the Pauli terms of pauli_terms variable into commuting sets.

    Uses algorithm defined in (Raeisi, Wiebe, Sanders, arXiv:1108.4318, 2011)
    to find commuting sets. Except uses commutation check from arXiv:1405.5749v2

    :param pauli_terms: A PauliSum object
    :returns: List of lists where each list contains a commuting set
    """
    m_terms = len(pauli_terms.terms)
    m_s = 1
    groups = []
    groups.append([pauli_terms.terms[0]])
    for j in range(1, m_terms):
        isAssigned_bool = False
        for p in range(m_s):  # check if it commutes with each group
            if isAssigned_bool is False:
                if check_commutation(groups[p], pauli_terms.terms[j]):
                    isAssigned_bool = True
                    groups[p].append(pauli_terms.terms[j])
        if isAssigned_bool is False:
            m_s += 1
            groups.append([pauli_terms.terms[j]])
    return groups


def is_identity(term: PauliDesignator) -> bool:
    """Return True if PauliTerm or PauliSum is a scalar multiple of identity, False otherwise.

    :param term: Either a PauliTerm or PauliSum
    :returns: True if the PauliTerm or PauliSum is a scalar multiple of identity, False otherwise
    """
    if isinstance(term, PauliTerm):
        return (len(term) == 0) and (not np.isclose(term.coefficient, 0))
    elif isinstance(term, PauliSum):
        return (len(term.terms) == 1) and (len(term.terms[0]) == 0) and (not np.isclose(term.terms[0].coefficient, 0))
    else:
        raise TypeError("is_identity only checks PauliTerms and PauliSum objects!")


def exponentiate(term: PauliTerm) -> Program:
    """Create a pyQuil program that simulates the unitary evolution exp(-1j * term).

    :param term: A pauli term to exponentiate
    :returns: A Program object
    """
    return exponential_map(term)(1.0)


def exponential_map(term: PauliTerm) -> Callable[[float], Program]:
    """Return a function f(alpha) that constructs the Program corresponding to exp(-1j*alpha*term).

    :param term: A pauli term to exponentiate
    :returns: A function that takes an angle parameter and returns a program.
    """
    if not isinstance(term.coefficient, Complex):
        raise TypeError(f"PauliTerm coefficient type must be complex, got {type(term.coefficient)}")

    if not np.isclose(np.imag(term.coefficient), 0.0):
        raise TypeError("PauliTerm coefficient must be real")

    coeff = term.coefficient.real
    term.coefficient = term.coefficient.real

    def exp_wrap(param: float) -> Program:
        prog = Program()
        if is_identity(term):
            prog.inst(X(0))
            prog.inst(PHASE(-param * coeff, 0))
            prog.inst(X(0))
            prog.inst(PHASE(-param * coeff, 0))
        elif is_zero(term):
            pass
        else:
            prog += _exponentiate_general_case(term, param)
        return prog

    return exp_wrap


def exponentiate_commuting_pauli_sum(
    pauli_sum: PauliSum,
) -> Callable[[float], Program]:
    """Return a function that maps all substituent PauliTerms and sums them into a program.

    NOTE: Use this function with care. Substituent PauliTerms should commute.

    :param pauli_sum: PauliSum to exponentiate.
    :returns: A function that parametrizes the exponential.
    """
    if not isinstance(pauli_sum, PauliSum):
        raise TypeError("Argument 'pauli_sum' must be a PauliSum.")

    fns = [exponential_map(term) for term in pauli_sum]

    def combined_exp_wrap(param: float) -> Program:
        return Program([f(param) for f in fns])

    return combined_exp_wrap


def exponentiate_pauli_sum(
    pauli_sum: Union[PauliSum, PauliTerm],
) -> NDArray[np.complex_]:
    r"""Exponentiate a sequence of PauliTerms, which may or may not commute.

    The Pauliterms must have fixed (non-parametric) coefficients. The coefficients are interpreted in cycles rather than
    radians or degrees.

    e^{-i\pi\sum_i \theta_i P_i}

    Thus, 1.0 corresponds to one full rotation.

    To produce a CZ gate:

    >>> from numpy import pi
    >>> phi = pi
    >>> coeff = phi / (-4 * pi)  # -0.25
    >>> hamiltonian = PauliTerm("Z", 0) * PauliTerm("Z", 1) - 1 * PauliTerm("Z", 0) - 1 * PauliTerm("Z", 1)
    >>> exponentiate_pauli_sum(coeff * hamiltonian)
    array([[...]])

    To produce the Quil XY(theta) gate, you can use:

    >>> theta = pi / 2
    >>> coeff = theta / (-4 * pi)
    >>> hamiltonian = PauliTerm("X", 0) * PauliTerm("X", 1) + PauliTerm("Y", 0) * PauliTerm("Y", 1)
    >>> exponentiate_pauli_sum(coeff * hamiltonian)
    array([[...]])

    A global phase is applied to the unitary such that the [0,0] element is always real.

    :param pauli_sum: PauliSum to exponentiate.
    :returns: The matrix exponent of the PauliSum
    """
    if isinstance(pauli_sum, PauliTerm):
        pauli_sum = PauliSum([pauli_sum])

    pauli_matrices = {
        "X": np.array([[0.0, 1.0], [1.0, 0.0]]),
        "Y": np.array([[0.0, 0.0 - 1.0j], [0.0 + 1.0j, 0.0]]),
        "Z": np.array([[1.0, 0.0], [0.0, -1.0]]),
        "I": np.array([[1.0, 0.0], [0.0, 1.0]]),
    }

    qubits = pauli_sum.get_qubits()
    for q in qubits:
        if not isinstance(q, int):
            raise ValueError("PauliSum must be defined on qubits, not placeholders.")
    qubits.sort()

    matrices = []
    for term in pauli_sum.terms:
        coeff = term.coefficient
        if not isinstance(coeff, Number):
            raise ValueError("PauliSum must have fixed coefficients.")
        qubit_paulis = {qubit: pauli for qubit, pauli in term.operations_as_set()}
        paulis = [qubit_paulis[q] if q in qubit_paulis else "I" for q in qubits]
        matrix = float(np.real(coeff)) * reduce(np.kron, [pauli_matrices[p] for p in paulis])
        matrices.append(matrix)
    generated_unitary = expm(-1j * np.pi * sum(matrices))
    phase = np.exp(-1j * np.angle(generated_unitary[0, 0]))
    return np.asarray(phase * generated_unitary, dtype=np.complex_)


def _exponentiate_general_case(pauli_term: PauliTerm, param: float) -> Program:
    """Return a Quil (Program()) object corresponding to the exponential of the pauli_term object.

    For example, exp[-1.0j * param * pauli_term]

    :param pauli_term: A PauliTerm to exponentiate
    :param param: scalar, non-complex, value
    :returns: A Quil program object
    """

    def reverse_hack(p: Program) -> Program:
        # A hack to produce a *temporary* program which reverses p.
        revp = Program()
        revp.inst(list(reversed(p.instructions)))
        return revp

    quil_prog = Program()
    change_to_z_basis = Program()
    change_to_original_basis = Program()
    cnot_seq = Program()
    prev_index = None
    highest_target_index = None

    for index, op in pauli_term:
        if not isinstance(index, (int, QubitPlaceholder)):
            raise ValueError("PauliTerm indices must be qubits or qubit placeholders, got {type(index)}.")
        if "X" == op:
            change_to_z_basis.inst(H(index))
            change_to_original_basis.inst(H(index))
        elif "Y" == op:
            change_to_z_basis.inst(RX(np.pi / 2.0, index))
            change_to_original_basis.inst(RX(-np.pi / 2.0, index))
        elif "I" == op:
            continue

        if prev_index is not None:
            cnot_seq.inst(CNOT(prev_index, index))

        prev_index = index
        highest_target_index = index

    # building rotation circuit
    quil_prog += change_to_z_basis
    quil_prog += cnot_seq
    if not isinstance(pauli_term.coefficient, Complex) or highest_target_index is None:
        raise ValueError("PauliTerm coefficient must be complex and highest_target_index must be set.")
    quil_prog.inst(RZ(2.0 * pauli_term.coefficient * param, highest_target_index))
    quil_prog += reverse_hack(cnot_seq)
    quil_prog += change_to_original_basis

    return quil_prog


def suzuki_trotter(trotter_order: int, trotter_steps: int) -> list[tuple[float, int]]:
    """Generate trotterization coefficients for a given number of Trotter steps.

    U = exp(A + B) is approximated as exp(w1*o1)exp(w2*o2)... This method returns
    a list [(w1, o1), (w2, o2), ... , (wm, om)] of tuples where o=0 corresponds
    to the A operator, o=1 corresponds to the B operator, and w is the
    coefficient in the exponential. For example, a second order Suzuki-Trotter
    approximation to exp(A + B) results in the following
    [(0.5/trotter_steps, 0), (1/trotter_steps, 1),
    (0.5/trotter_steps, 0)] * trotter_steps.

    :param trotter_order: order of Suzuki-Trotter approximation
    :param trotter_steps: number of steps in the approximation
    :returns: List of tuples corresponding to the coefficient and operator
              type: o=0 is A and o=1 is B.
    """
    p1 = p2 = p4 = p5 = 1.0 / (4 - (4 ** (1.0 / 3)))
    p3 = 1 - 4 * p1
    trotter_dict: dict[int, list[tuple[float, int]]] = {
        1: [(1, 0), (1, 1)],
        2: [(0.5, 0), (1, 1), (0.5, 0)],
        3: [
            (7.0 / 24, 0),
            (2.0 / 3.0, 1),
            (3.0 / 4.0, 0),
            (-2.0 / 3.0, 1),
            (-1.0 / 24, 0),
            (1.0, 1),
        ],
        4: [
            (p5 / 2, 0),
            (p5, 1),
            (p5 / 2, 0),
            (p4 / 2, 0),
            (p4, 1),
            (p4 / 2, 0),
            (p3 / 2, 0),
            (p3, 1),
            (p3 / 2, 0),
            (p2 / 2, 0),
            (p2, 1),
            (p2 / 2, 0),
            (p1 / 2, 0),
            (p1, 1),
            (p1 / 2, 0),
        ],
    }

    order_slices = [(x0 / trotter_steps, x1) for x0, x1 in trotter_dict[trotter_order]]
    order_slices = order_slices * trotter_steps
    return order_slices


def is_zero(pauli_object: PauliDesignator) -> bool:
    """Return True if a PauliTerm or PauliSum is zero, False otherwise.

    :param pauli_object: Either a PauliTerm or PauliSum
    :returns: True if PauliTerm is zero, False otherwise
    """
    if isinstance(pauli_object, PauliTerm):
        if not isinstance(pauli_object.coefficient, Complex):
            raise ValueError("PauliTerm coefficient must be fixed and complex.")
        return bool(np.isclose(pauli_object.coefficient, 0))
    elif isinstance(pauli_object, PauliSum):
        if not isinstance(pauli_object.terms[0].coefficient, Complex):
            raise ValueError("PauliTerm coefficient must be fixed and complex.")
        return len(pauli_object.terms) == 1 and bool(np.isclose(pauli_object.terms[0].coefficient, 0))
    else:
        raise TypeError("is_zero only checks PauliTerms and PauliSum objects!")


def trotterize(
    first_pauli_term: PauliTerm,
    second_pauli_term: PauliTerm,
    trotter_order: int = 1,
    trotter_steps: int = 1,
) -> Program:
    """Create a Quil program that approximates exp( (A + B)t) where A and B are PauliTerm operators.

    :param first_pauli_term: PauliTerm denoted `A`
    :param second_pauli_term: PauliTerm denoted `B`
    :param trotter_order: Optional argument indicating the Suzuki-Trotter
                          approximation order--only accepts orders 1, 2, 3, 4.
    :param trotter_steps: Optional argument indicating the number of products
                          to decompose the exponential into.

    :return: Quil program
    """
    if not (1 <= trotter_order < 5):
        raise ValueError("trotterize only accepts trotter_order in {1, 2, 3, 4}.")

    commutator = (first_pauli_term * second_pauli_term) + (-1 * second_pauli_term * first_pauli_term)

    prog = Program()
    if is_zero(commutator):
        param_exp_prog_one = exponential_map(first_pauli_term)
        exp_prog = param_exp_prog_one(1)
        prog += exp_prog
        param_exp_prog_two = exponential_map(second_pauli_term)
        exp_prog = param_exp_prog_two(1)
        prog += exp_prog
        return prog

    order_slices = suzuki_trotter(trotter_order, trotter_steps)
    for coeff, operator in order_slices:
        if operator == 0:
            param_prog = exponential_map(coeff * first_pauli_term)
            exp_prog = param_prog(1)
            prog += exp_prog
        else:
            param_prog = exponential_map(coeff * second_pauli_term)
            exp_prog = param_prog(1)
            prog += exp_prog
    return prog



"""Convert a QCS Instruction Set Architecture to a NetworkX graph."""

import networkx as nx
from qcs_sdk.qpu.isa import InstructionSetArchitecture


def qcs_isa_to_graph(isa: InstructionSetArchitecture) -> nx.Graph:
    """Convert a QCS Instruction Set Architecture to a NetworkX graph."""
    return nx.from_edgelist(edge.node_ids for edge in isa.architecture.edges)



"""An implementation of an AbstractQuantumProcessor initialized with a user constructed ``CompilerISA``."""

import networkx as nx

from pyquil.external.rpcq import CompilerISA
from pyquil.quantum_processor._base import AbstractQuantumProcessor
from pyquil.quantum_processor.transformers import compiler_isa_to_graph


class CompilerQuantumProcessor(AbstractQuantumProcessor):
    """An AbstractQuantumProcessor initialized with a user constructed ``CompilerISA``."""

    _isa: CompilerISA

    def __init__(self, isa: CompilerISA) -> None:
        """Initialize a CompilerQuantumProcessor with a given ``CompilerISA``."""
        self._isa = isa

    def qubit_topology(self) -> nx.Graph:
        """Return a NetworkX graph that represents the connectivity of qubits in this quantum_processor."""
        return compiler_isa_to_graph(self._isa)

    def to_compiler_isa(self) -> CompilerISA:
        """Return the CompilerISA that this quantum_processor is initialized with."""
        return self._isa

    def qubits(self) -> list[int]:
        """Return the qubits in the quantum_processor topology as a sorted list."""
        return sorted([int(node_id) for node_id, node in self._isa.qubits.items()])



"""Various transformers for quantum processor ISA representations."""

__all__ = [
    "compiler_isa_to_graph",
    "graph_to_compiler_isa",
    "QCSISAParseError",
    "qcs_isa_to_compiler_isa",
    "qcs_isa_to_graph",
    "GraphGateError",
]

from pyquil.quantum_processor.transformers.compiler_isa_to_graph import compiler_isa_to_graph
from pyquil.quantum_processor.transformers.graph_to_compiler_isa import GraphGateError, graph_to_compiler_isa
from pyquil.quantum_processor.transformers.qcs_isa_to_compiler_isa import (
    QCSISAParseError,
    qcs_isa_to_compiler_isa,
)
from pyquil.quantum_processor.transformers.qcs_isa_to_graph import qcs_isa_to_graph



##############################################################################
# Copyright 2017-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""Classes that represent the atomic building blocks of Quil expressions."""

import inspect
from collections.abc import Iterable, Mapping, Sequence
from fractions import Fraction
from numbers import Number
from typing import (
    Any,
    Callable,
    ClassVar,
    NoReturn,
    Optional,
    Union,
    cast,
)

import numpy as np
import quil.expression as quil_rs_expr
import quil.instructions as quil_rs
from deprecated.sphinx import deprecated
from typing_extensions import Self


class QuilAtom:
    """Abstract class for atomic elements of Quil."""

    def out(self) -> str:
        """Return the element as a valid Quil string."""
        raise NotImplementedError()

    def __str__(self) -> str:
        """Get a string representation of the element, possibly not valid Quil."""
        raise NotImplementedError()

    def __eq__(self, other: object) -> bool:
        """Return True if the other object is equal to this one."""
        raise NotImplementedError()

    def __hash__(self) -> int:
        """Return a hash of the object."""
        raise NotImplementedError()


class Qubit(QuilAtom):
    """Representation of a qubit.

    :param index: Index of the qubit.
    """

    def __init__(self, index: int):
        """Initialize a qubit."""
        if not (isinstance(index, int) and index >= 0):
            raise TypeError("Addr index must be a non-negative int")
        self.index = index

    def out(self) -> str:
        """Return the element as a valid Quil string."""
        return str(self.index)

    def __str__(self) -> str:
        return str(self.index)

    def __repr__(self) -> str:
        return f"<Qubit {self.index}>"

    def __hash__(self) -> int:
        return hash(self.index)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Qubit) and other.index == self.index


class FormalArgument(QuilAtom):
    """Representation of a formal argument associated with a DEFCIRCUIT or DEFGATE ... AS PAULI-SUM or DEFCAL form."""

    def __init__(self, name: str):
        """Initialize a formal argument."""
        if not isinstance(name, str):
            raise TypeError("Formal arguments must be named by a string.")
        self.name = name

    def out(self) -> str:
        """Return the element as a valid Quil string."""
        return str(self)

    @property
    @deprecated(
        "Getting the index of a FormalArgument is invalid. This method will be removed in a future release.",
        version="4.0",
    )
    def index(self) -> NoReturn:
        """Formal arguments do not have an index. Using this property raises a RuntimeError."""
        raise RuntimeError(f"Cannot derive an index from a FormalArgument {self}")

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"<FormalArgument {self.name}>"

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, FormalArgument) and other.name == self.name


class QubitPlaceholder(QuilAtom):
    """A placeholder for a qubit.

    This is useful for constructing circuits without assigning qubits to specific indices.
    Qubit placeholders must be resolved to actual qubits before they can be used in a program.
    """

    def __init__(self, placeholder: Optional[quil_rs.QubitPlaceholder] = None):
        """Initialize a qubit placeholder, or get a new handle for an existing placeholder."""
        if placeholder is not None:
            self._placeholder = placeholder
        else:
            self._placeholder = quil_rs.QubitPlaceholder()

    @staticmethod
    def register(n: int) -> list["QubitPlaceholder"]:
        r"""Return a 'register' of ``n`` QubitPlaceholders.

        >>> from pyquil import Program
        >>> from pyquil.gates import H
        >>> from pyquil.quil import address_qubits
        >>> from pyquil.quilatom import QubitPlaceholder
        >>> qs = QubitPlaceholder.register(8)  # a qubyte
        >>> prog = Program(H(q) for q in qs)
        >>> address_qubits(prog).out()
        'H 0\nH 1\nH 2\nH 3\nH 4\nH 5\nH 6\nH 7\n'
        >>>

        The returned register is a Python list of QubitPlaceholder objects, so all
        normal list semantics apply.

        :param n: The number of qubits in the register
        """
        return [QubitPlaceholder() for _ in range(n)]

    def out(self) -> str:
        """Raise a RuntimeError, as Qubit placeholders are not valid Quil."""
        raise RuntimeError(f"Qubit {self} has not been assigned an index")

    @property
    def index(self) -> NoReturn:
        """Raise a RuntimeError, as Qubit placeholders do not have an index."""
        raise RuntimeError(f"Qubit {self} has not been assigned an index")

    def __str__(self) -> str:
        return f"q{id(self)}"

    def __repr__(self) -> str:
        return f"q{id(self)}"

    def __hash__(self) -> int:
        return hash(self._placeholder)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, quil_rs.QubitPlaceholder):
            return self._placeholder == other
        if isinstance(other, QubitPlaceholder):
            return self._placeholder == other._placeholder
        return False

    def __lt__(self, other: object) -> bool:
        if isinstance(other, quil_rs.QubitPlaceholder):
            return self._placeholder < other
        if isinstance(other, QubitPlaceholder):
            return self._placeholder < other._placeholder
        raise TypeError(f"Comparison between LabelPlaceholder and {type(other)} is not supported.")


QubitDesignator = Union[Qubit, QubitPlaceholder, FormalArgument, int]


def _convert_to_rs_qubit(qubit: Union[QubitDesignator, quil_rs.Qubit, QubitPlaceholder]) -> quil_rs.Qubit:
    if isinstance(qubit, quil_rs.Qubit):
        return qubit
    if isinstance(qubit, Qubit):
        return quil_rs.Qubit.from_fixed(qubit.index)
    if isinstance(qubit, QubitPlaceholder):
        return quil_rs.Qubit.from_placeholder(qubit._placeholder)
    if isinstance(qubit, FormalArgument):
        return quil_rs.Qubit.from_variable(qubit.name)
    if isinstance(qubit, int):
        return quil_rs.Qubit.from_fixed(qubit)
    raise ValueError(f"{type(qubit)} is not a valid QubitDesignator")


def _convert_to_rs_qubits(qubits: Iterable[QubitDesignator]) -> list[quil_rs.Qubit]:
    return [_convert_to_rs_qubit(qubit) for qubit in qubits]


def _convert_to_py_qubit(qubit: Union[QubitDesignator, quil_rs.Qubit, quil_rs.QubitPlaceholder]) -> QubitDesignator:
    if isinstance(qubit, quil_rs.Qubit):
        if qubit.is_fixed():
            return Qubit(qubit.to_fixed())
        if qubit.is_variable():
            return FormalArgument(qubit.to_variable())
        if qubit.is_placeholder():
            return QubitPlaceholder(placeholder=qubit.to_placeholder())
    if isinstance(qubit, (Qubit, QubitPlaceholder, FormalArgument, Parameter, int)):
        return qubit
    raise ValueError(f"{type(qubit)} is not a valid QubitDesignator")


def _convert_to_py_qubits(qubits: Iterable[Union[QubitDesignator, quil_rs.Qubit]]) -> list[QubitDesignator]:
    return [_convert_to_py_qubit(qubit) for qubit in qubits]


def unpack_qubit(qubit: Union[QubitDesignator, FormalArgument]) -> Union[Qubit, QubitPlaceholder, FormalArgument]:
    """Get a qubit from an object.

    :param qubit: the qubit designator to unpack.
    :return: A Qubit or QubitPlaceholder instance
    """
    if isinstance(qubit, int):
        return Qubit(qubit)
    elif isinstance(qubit, Qubit):
        return qubit
    elif isinstance(qubit, QubitPlaceholder):
        return qubit
    elif isinstance(qubit, FormalArgument):
        return qubit
    else:
        raise TypeError("qubit should be an int or Qubit or QubitPlaceholder instance")


def qubit_index(qubit: QubitDesignator) -> int:
    """Get the index of a QubitDesignator.

    :param qubit: the qubit designator.
    :return: An int that is the qubit index.
    """
    if isinstance(qubit, Qubit):
        return qubit.index
    elif isinstance(qubit, int):
        return qubit
    else:
        raise TypeError(f"Cannot unwrap unaddressed QubitPlaceholder: {qubit}")


# Like the Tuple, the List must be length 2, where the first item is a string and the second an
# int. However, specifying Union[str, int] as the generic type argument to List doesn't sufficiently
# constrain the types, and mypy gets confused in unpack_classical_reg, below. Hence, just specify
# list[Any] here.
MemoryReferenceDesignator = Union["MemoryReference", quil_rs.MemoryReference, tuple[str, int], list[Any], str]


def unpack_classical_reg(c: MemoryReferenceDesignator) -> "MemoryReference":
    """Get the address for a classical register.

    :param c: A list of length 2, a pair, a string (to be interpreted as name[0]), or a
        MemoryReference.
    :return: The address as a MemoryReference.
    """
    if isinstance(c, list) or isinstance(c, tuple):
        if len(c) > 2 or len(c) == 0:
            raise ValueError("if c is a list/tuple, it should be of length <= 2")
        if len(c) == 1:
            c = (c[0], 0)
        if not isinstance(c[0], str):
            raise ValueError("if c is a list/tuple, its first member should be a string")
        if not isinstance(c[1], int):
            raise ValueError("if c is a list/tuple, its second member should be an int")
        return MemoryReference(c[0], c[1])
    if isinstance(c, MemoryReference):
        return c
    elif isinstance(c, str):
        return MemoryReference(c, 0)
    else:
        raise TypeError("c should be a list of length 2, a pair, a string, or a MemoryReference")


class Label(QuilAtom):
    """Representation of a label.

    :param label_name: The label name.
    """

    def __init__(self, label_name: str):
        """Initialize a new label."""
        self.target = quil_rs.Target.from_fixed(label_name)

    @staticmethod
    def _from_rs_target(target: quil_rs.Target) -> "Label":
        return Label(target.to_fixed())

    def out(self) -> str:
        """Return the label as a valid Quil string."""
        return self.target.to_quil()

    @property
    def name(self) -> str:
        """Return the label name."""
        return self.target.to_fixed()

    @name.setter
    def name(self, label_name: str) -> None:
        self.target = quil_rs.Target.from_fixed(label_name)

    def __str__(self) -> str:
        return self.target.to_quil_or_debug()

    def __repr__(self) -> str:
        return repr(self.target)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Label):
            return self.target == other.target
        return False

    def __hash__(self) -> int:
        return hash(self.target)


class LabelPlaceholder(QuilAtom):
    """A placeholder for a Quil label.

    This is useful for constructing circuits without needing to name them.
    All placeholders must be resolved to actual labels before they can be used in a program.
    """

    def __init__(self, prefix: str = "L", *, placeholder: Optional[quil_rs.TargetPlaceholder] = None):
        """Initialize a new label placeholder."""
        if placeholder:
            self.target = quil_rs.Target.from_placeholder(placeholder)
        else:
            self.target = quil_rs.Target.from_placeholder(quil_rs.TargetPlaceholder(prefix))

    @staticmethod
    def _from_rs_target(target: quil_rs.Target) -> "LabelPlaceholder":
        return LabelPlaceholder(placeholder=target.to_placeholder())

    @property
    def prefix(self) -> str:
        """Get the prefix of the label placeholder."""
        return self.target.to_placeholder().base_label

    def out(self) -> str:
        """Raise a RuntimeError, as label placeholders are not valid Quil."""
        raise RuntimeError("Label has not been assigned a name")

    def __str__(self) -> str:
        return self.target.to_quil_or_debug()

    def __repr__(self) -> str:
        return repr(self.target)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, LabelPlaceholder):
            return self.target == other.target
        return False

    def __hash__(self) -> int:
        return hash(self.target)


ParameterDesignator = Union["Expression", "MemoryReference", int, float, complex, np.number]


def _convert_to_rs_expression(
    parameter: Union[ParameterDesignator, quil_rs_expr.Expression],
) -> quil_rs_expr.Expression:
    if isinstance(parameter, quil_rs_expr.Expression):
        return parameter
    elif isinstance(parameter, (int, float, complex, np.number)):
        return quil_rs_expr.Expression.from_number(complex(parameter))
    elif isinstance(parameter, (Expression, MemoryReference)):
        return quil_rs_expr.Expression.parse(str(parameter))
    raise ValueError(f"{type(parameter)} is not a valid ParameterDesignator")


def _convert_to_rs_expressions(
    parameters: Sequence[Union[ParameterDesignator, quil_rs_expr.Expression]],
) -> list[quil_rs_expr.Expression]:
    return [_convert_to_rs_expression(parameter) for parameter in parameters]


@deprecated(version="4.0", reason="This function has been superseded by the `quil` package and will be removed soon.")
def format_parameter(element: ParameterDesignator) -> str:
    """Format a particular parameter.

    Essentially the same as built-in formatting except using 'i' instead of 'j' for the imaginary number.

    :param element: The parameter to format for Quil output.
    """
    if isinstance(element, (int, np.integer)):
        return repr(element)
    elif isinstance(element, float):
        return _check_for_pi(element)
    elif isinstance(element, complex):
        out = ""
        r = element.real
        i = element.imag
        if i == 0:
            return repr(r)

        if r != 0:
            out += repr(r)

        if i == 1:
            if not np.isclose(r, 0, atol=1e-14):
                raise ValueError(f"Invalid complex number: {element}")
            out = "i"
        elif i == -1:
            if not np.isclose(r, 0, atol=1e-14):
                raise ValueError(f"Invalid complex number: {element}")
            out = "-i"
        elif i < 0:
            out += repr(i) + "i"
        elif r != 0:
            out += "+" + repr(i) + "i"
        else:
            out += repr(i) + "i"

        return out
    elif isinstance(element, MemoryReference):
        return str(element)
    elif isinstance(element, Expression):
        return _expression_to_string(element)
    raise AssertionError(f"Invalid parameter: {element}")


ExpressionValueDesignator = Union[int, float, complex]
ExpressionDesignator = Union["Expression", ExpressionValueDesignator]


def _convert_to_py_expression(
    expression: Union[
        ParameterDesignator,
        ExpressionDesignator,
        ExpressionValueDesignator,
        quil_rs_expr.Expression,
        quil_rs.MemoryReference,
    ],
) -> ExpressionDesignator:
    if isinstance(expression, (Expression, Number)):
        return expression
    if isinstance(expression, quil_rs_expr.Expression):
        if expression.is_pi():
            return np.pi
        if expression.is_number():
            return expression.to_number()
        if expression.is_variable():
            return Parameter(expression.to_variable())
        if expression.is_infix():
            return BinaryExp._from_rs_infix_expression(expression.to_infix())
        if expression.is_address():
            return MemoryReference._from_rs_memory_reference(expression.to_address())
        if expression.is_function_call():
            fc = expression.to_function_call()
            parameter = _convert_to_py_expression(fc.expression)
            if fc.function == quil_rs_expr.ExpressionFunction.Cis:
                return quil_cis(parameter)
            if fc.function == quil_rs_expr.ExpressionFunction.Cosine:
                return quil_cos(parameter)
            if fc.function == quil_rs_expr.ExpressionFunction.Exponent:
                return quil_exp(parameter)
            if fc.function == quil_rs_expr.ExpressionFunction.Sine:
                return quil_sin(parameter)
            if fc.function == quil_rs_expr.ExpressionFunction.SquareRoot:
                return quil_sqrt(parameter)
        if expression.is_prefix():
            prefix = expression.to_prefix()
            py_expression = _convert_to_py_expression(prefix.expression)
            if prefix == quil_rs_expr.PrefixOperator.Plus:
                return py_expression
            elif isinstance(py_expression, (int, float, complex, Expression)):
                return -py_expression
    raise TypeError(f"{type(expression)} is not a valid ExpressionDesignator")


def _convert_to_py_expressions(
    expressions: Sequence[
        Union[ParameterDesignator, ExpressionDesignator, quil_rs_expr.Expression, quil_rs.MemoryReference]
    ],
) -> Sequence[ExpressionDesignator]:
    return [_convert_to_py_expression(expression) for expression in expressions]


class Expression:
    """Expression involving some unbound parameters.

    Parameters in Quil are represented as a label like '%x' for the parameter named 'x'. An example expression therefore
    may be '%x*(%y/4)'.

    Expressions may also have function calls, supported functions in Quil are sin, cos, sqrt, exp,
    and cis.

    This class overrides all the Python operators that are supported by Quil.
    """

    def __str__(self) -> str:
        return _expression_to_string(self)

    def __repr__(self) -> str:
        return str(self.__class__.__name__) + "(" + ",".join(map(repr, self.__dict__.values())) + ")"

    def __add__(self, other: ExpressionDesignator) -> "Add":
        return Add(self, other)

    def __radd__(self, other: ExpressionDesignator) -> "Add":
        return Add(other, self)

    def __sub__(self, other: ExpressionDesignator) -> "Sub":
        return Sub(self, other)

    def __rsub__(self, other: ExpressionDesignator) -> "Sub":
        return Sub(other, self)

    def __mul__(self, other: ExpressionDesignator) -> "Mul":
        return Mul(self, other)

    def __rmul__(self, other: ExpressionDesignator) -> "Mul":
        return Mul(other, self)

    def __div__(self, other: ExpressionDesignator) -> "Div":
        return Div(self, other)

    __truediv__ = __div__

    def __rdiv__(self, other: ExpressionDesignator) -> "Div":
        return Div(other, self)

    __rtruediv__ = __rdiv__

    def __pow__(self, other: ExpressionDesignator) -> "Pow":
        return Pow(self, other)

    def __rpow__(self, other: ExpressionDesignator) -> "Pow":
        return Pow(other, self)

    def __neg__(self) -> "Mul":
        return Mul(-1, self)

    def _substitute(self, d: Any) -> ExpressionDesignator:
        return self

    def _evaluate(self) -> np.complex128:
        """Attempt to evaluate the expression to by simplifying it to a complex number.

        Expression simplification can be slow, especially for large recursive expressions.
        This method will raise a ValueError if the expression cannot be simplified to a complex
        number.
        """
        expr = quil_rs_expr.Expression.parse(str(self))
        expr.simplify()  # type: ignore[no-untyped-call]
        if not expr.is_number():
            raise ValueError(f"Cannot evaluate expression {self} to a number. Got {expr}.")
        return np.complex128(expr.to_number())

    def __float__(self) -> float:
        """Return a copy of the expression as a float by attempting to simplify the expression.

        Expression simplification can be slow, especially for large recursive expressions.
        This cast will raise a ValueError if simplification doesn't result in a real number.
        """
        value = self._evaluate()
        if value.imag != 0:
            raise ValueError(f"Cannot convert complex value with non-zero imaginary value to float: {value}")
        return float(value.real)

    def __array__(self, dtype: Optional[np.dtype] = None) -> np.ndarray:
        """Implement the numpy array protocol for this expression.

        If the dtype is not object, then there will be an attempt to simplify the expression to a
        complex number. If the expression cannot be simplified to one, then fallback to the
        object representation of the expression.

        Note that expression simplification can be slow for large recursive expressions.
        """
        try:
            if dtype != object:
                return np.asarray(self._evaluate(), dtype=dtype)
            raise ValueError
        except ValueError:
            # np.asarray(self, ...) would cause an infinite recursion error, so we build the array with a
            # placeholder value, then replace it with self after.
            array = np.asarray(None, dtype=object)
            array.flat[0] = self
            return array


ParameterSubstitutionsMapDesignator = Mapping[Union["Parameter", "MemoryReference"], ExpressionValueDesignator]


def substitute(expr: ExpressionDesignator, d: ParameterSubstitutionsMapDesignator) -> ExpressionDesignator:
    r"""Explicitly evaluate as much of ``expr`` as possible, using substitutions from `d`.

    This supports substitution of both parameters and memory references. Each memory reference must be individually
    assigned a value at each memory offset to be substituted.

    :param expr: The expression whose parameters or memory references are to be substituted.
    :param d: Numerical substitutions for parameters or memory references.
    :return: A partially simplified Expression, or a number.
    """
    if isinstance(expr, Expression):
        return expr._substitute(d)
    return expr


def substitute_array(a: Union[Sequence[Expression], np.ndarray], d: ParameterSubstitutionsMapDesignator) -> np.ndarray:
    """Apply ``substitute`` to all elements of an array ``a`` and return the resulting array.

    :param a: The array of expressions whose parameters or memory references are to be substituted.
    :param d: Numerical substitutions for parameters or memory references, for all array elements.
    :return: An array of partially substituted Expressions, or numbers.
    """
    a = np.asarray(a, order="C")
    return np.array([substitute(v, d) for v in a.flat]).reshape(a.shape)


class Parameter(QuilAtom, Expression):
    """Parameters in Quil are represented as a label like '%x' for the parameter named 'x'."""

    def __init__(self, name: str):
        """Initialize a new parameter."""
        self.name = name

    def out(self) -> str:
        """Return the parameter as a valid Quil string."""
        return "%" + self.name

    def _substitute(self, d: ParameterSubstitutionsMapDesignator) -> Union["Parameter", ExpressionValueDesignator]:
        return d.get(self, self)

    def __str__(self) -> str:
        return "%" + self.name

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Parameter) and other.name == self.name


class Function(Expression):
    """Base class for standard Quil functions."""

    def __init__(
        self,
        name: str,
        expression: ExpressionDesignator,
        fn: Callable[[ExpressionValueDesignator], ExpressionValueDesignator],
    ):
        """Initialize a new function."""
        self.name = name
        self.expression = expression
        self.fn = fn

    @classmethod
    def _from_rs_function_call(cls, function_call: quil_rs_expr.FunctionCallExpression) -> "Function":
        expression = _convert_to_py_expression(function_call.expression)
        if function_call.function == quil_rs_expr.ExpressionFunction.Cis:
            return quil_cis(expression)
        if function_call.function == quil_rs_expr.ExpressionFunction.Cosine:
            return quil_cos(expression)
        if function_call.function == quil_rs_expr.ExpressionFunction.Exponent:
            return quil_exp(expression)
        if function_call.function == quil_rs_expr.ExpressionFunction.Sine:
            return quil_sin(expression)
        return quil_sqrt(expression)

    def _substitute(self, d: ParameterSubstitutionsMapDesignator) -> Union["Function", ExpressionValueDesignator]:
        sop = substitute(self.expression, d)
        if isinstance(sop, Expression):
            return Function(self.name, sop, self.fn)
        return self.fn(sop)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Function) and self.name == other.name and self.expression == other.expression

    def __neq__(self, other: "Function") -> bool:
        return not self.__eq__(other)


def quil_sin(expression: ExpressionDesignator) -> Function:
    """Quil COS() function."""
    return Function("SIN", expression, np.sin)


def quil_cos(expression: ExpressionDesignator) -> Function:
    """Quil SIN() function."""
    return Function("COS", expression, np.cos)


def quil_sqrt(expression: ExpressionDesignator) -> Function:
    """Quil SQRT() function."""
    return Function("SQRT", expression, np.sqrt)


def quil_exp(expression: ExpressionDesignator) -> Function:
    """Quil EXP() function."""
    return Function("EXP", expression, np.exp)


def quil_cis(expression: ExpressionDesignator) -> Function:
    """Quil CIS() function."""

    def _cis(x: ExpressionValueDesignator) -> complex:
        # numpy doesn't ship with type stubs, so mypy assumes anything coming from numpy has type
        # Any, hence we need to cast the return type to complex here to satisfy the type checker.
        return cast(complex, np.exp(1j * x))

    return Function("CIS", expression, _cis)


class BinaryExp(Expression):
    """A Quil binary expression."""

    operator: ClassVar[str]
    precedence: ClassVar[int]
    associates: ClassVar[str]

    @staticmethod
    def fn(a: ExpressionDesignator, b: ExpressionDesignator) -> Union["BinaryExp", ExpressionValueDesignator]:
        """Perform the operation on the two expressions."""
        raise NotImplementedError

    def __init__(self, op1: ExpressionDesignator, op2: ExpressionDesignator):
        """Initialize a new binary expression."""
        self.op1 = op1
        self.op2 = op2

    @classmethod
    def _from_rs_infix_expression(
        cls, infix_expression: quil_rs_expr.InfixExpression
    ) -> Union["BinaryExp", ExpressionValueDesignator]:
        left = _convert_to_py_expression(infix_expression.left)
        right = _convert_to_py_expression(infix_expression.right)
        if infix_expression.operator == quil_rs_expr.InfixOperator.Plus:
            return Add.fn(left, right)
        if infix_expression.operator == quil_rs_expr.InfixOperator.Minus:
            return Sub.fn(left, right)
        if infix_expression.operator == quil_rs_expr.InfixOperator.Slash:
            return Div.fn(left, right)
        if infix_expression.operator == quil_rs_expr.InfixOperator.Star:
            return Mul.fn(left, right)
        if infix_expression.operator == quil_rs_expr.InfixOperator.Caret:
            return Pow.fn(left, right)
        raise ValueError(f"{type(infix_expression)} is not a valid InfixExpression")

    def _substitute(self, d: ParameterSubstitutionsMapDesignator) -> Union["BinaryExp", ExpressionValueDesignator]:
        sop1, sop2 = substitute(self.op1, d), substitute(self.op2, d)
        return self.fn(sop1, sop2)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and self.op1 == other.op1 and self.op2 == other.op2

    def __neq__(self, other: "BinaryExp") -> bool:
        return not self.__eq__(other)


class Add(BinaryExp):
    """The addition operation."""

    operator = " + "
    precedence = 1
    associates = "both"

    @staticmethod
    def fn(a: ExpressionDesignator, b: ExpressionDesignator) -> Union["Add", ExpressionValueDesignator]:
        """Perform the addition operation."""
        return a + b

    def __init__(self, op1: ExpressionDesignator, op2: ExpressionDesignator):
        """Initialize a new addition operation between two expressions."""
        super().__init__(op1, op2)


class Sub(BinaryExp):
    """The subtraction operation."""

    operator = " - "
    precedence = 1
    associates = "left"

    @staticmethod
    def fn(a: ExpressionDesignator, b: ExpressionDesignator) -> Union["Sub", ExpressionValueDesignator]:
        """Perform the subtraction operation."""
        return a - b

    def __init__(self, op1: ExpressionDesignator, op2: ExpressionDesignator):
        """Initialize a new addition operation between two expressions."""
        super().__init__(op1, op2)


class Mul(BinaryExp):
    """The multiplication operation."""

    operator = "*"
    precedence = 2
    associates = "both"

    @staticmethod
    def fn(a: ExpressionDesignator, b: ExpressionDesignator) -> Union["Mul", ExpressionValueDesignator]:
        """Perform the multiplication operation."""
        return a * b

    def __init__(self, op1: ExpressionDesignator, op2: ExpressionDesignator):
        """Initialize a new multiplication operation between two expressions."""
        super().__init__(op1, op2)


class Div(BinaryExp):
    """The division operation."""

    operator = "/"
    precedence = 2
    associates = "left"

    @staticmethod
    def fn(a: ExpressionDesignator, b: ExpressionDesignator) -> Union["Div", ExpressionValueDesignator]:
        """Perform the division operation."""
        return a / b

    def __init__(self, op1: ExpressionDesignator, op2: ExpressionDesignator):
        """Initialize a new division operation between two expressions."""
        super().__init__(op1, op2)


class Pow(BinaryExp):
    """The exponentiation operation."""

    operator = "^"
    precedence = 3
    associates = "right"

    @staticmethod
    def fn(a: ExpressionDesignator, b: ExpressionDesignator) -> Union["Pow", ExpressionValueDesignator]:
        """Perform the exponentiation operation."""
        return a**b

    def __init__(self, op1: ExpressionDesignator, op2: ExpressionDesignator):
        """Initialize a new exponentiation operation between two expressions."""
        super().__init__(op1, op2)


def _expression_to_string(expression: ExpressionDesignator) -> str:
    """Convert an expression to a string.

    This operation is recursive, and takes into account precedence and associativity when placing parenthesis.

    :param expression: expression involving parameters
    :return: string such as '%x*(%y-4)'
    """
    if isinstance(expression, BinaryExp):
        left = _expression_to_string(expression.op1)
        if isinstance(expression.op1, BinaryExp) and not (
            expression.op1.precedence > expression.precedence
            or expression.op1.precedence == expression.precedence
            and expression.associates in ("left", "both")
        ):
            left = "(" + left + ")"

        right = _expression_to_string(expression.op2)
        if isinstance(expression.op2, BinaryExp) and not (
            expression.precedence < expression.op2.precedence
            or expression.precedence == expression.op2.precedence
            and expression.associates in ("right", "both")
        ):
            right = "(" + right + ")"
        # If op2 is a float, it will maybe represented as a multiple
        # of pi in right. If that is the case, then we need to take
        # extra care to insert parens. Similarly, complex numbers need
        # to be wrapped in parens so the imaginary part is captured.
        # See gh-943,1734.
        elif (isinstance(expression.op2, float) and ("pi" in right and right != "pi")) or isinstance(
            expression.op2, complex
        ):
            right = "(" + right + ")"

        return left + expression.operator + right
    elif isinstance(expression, Function):
        return expression.name + "(" + _expression_to_string(expression.expression) + ")"
    elif isinstance(expression, Parameter):
        return str(expression)
    else:
        return format_parameter(expression)


def _contained_parameters(expression: ExpressionDesignator) -> set[Parameter]:
    """Determine which parameters are contained in this expression.

    :param expression: expression involving parameters
    :return: set of parameters contained in this expression
    """
    if isinstance(expression, BinaryExp):
        return _contained_parameters(expression.op1) | _contained_parameters(expression.op2)
    elif isinstance(expression, Function):
        return _contained_parameters(expression.expression)
    elif isinstance(expression, Parameter):
        return {expression}
    else:
        return set()


def _check_for_pi(element: float) -> str:
    """Return the float as a string, expressing the float as a multiple of pi if possible.

    More specifically, check to see if there exists a rational number r = p/q in reduced form for which the difference
    between element/np.pi and r is small and q <= 8.

    :param element: the number to check
    :return element: pretty print string if true, else standard representation.
    """
    frac = Fraction(element / np.pi).limit_denominator(8)
    num, den = frac.numerator, frac.denominator
    sign = "-" if num < 0 else ""
    if num / float(den) == element / np.pi:
        if num == 0:
            return "0"
        elif abs(num) == 1 and den == 1:
            return sign + "pi"
        elif abs(num) == 1:
            return sign + "pi/" + repr(den)
        elif den == 1:
            return repr(num) + "*pi"
        else:
            return repr(num) + "*pi/" + repr(den)
    else:
        return repr(element)


class MemoryReference(QuilAtom, Expression):
    """Representation of a reference to a classical memory address.

    :param name: The name of the variable
    :param offset: Everything in Quil is a C-style array, so every memory reference has an offset.
    :param declared_size: The optional size of the named declaration. This can be used for bounds
        checking, but isn't. It is used for pretty-printing to quil by deciding whether to output
        memory references with offset 0 as either e.g. ``ro[0]`` or ``beta`` depending on whether
        the declared variable is of length >1 or 1, resp.
    """

    def __init__(self, name: str, offset: int = 0, declared_size: Optional[int] = None):
        """Initialize a new memory reference."""
        if not isinstance(offset, int) or offset < 0:
            raise TypeError("MemoryReference offset must be a non-negative int")
        self.name = name
        self.offset = offset
        self.declared_size = declared_size

    @classmethod
    def _from_rs_memory_reference(cls, memory_reference: quil_rs.MemoryReference) -> "MemoryReference":
        return cls(memory_reference.name, memory_reference.index)

    @classmethod
    def _from_parameter_str(cls, memory_reference_str: str) -> "MemoryReference":
        expression = quil_rs_expr.Expression.parse(memory_reference_str)
        if expression.is_address():
            return cls._from_rs_memory_reference(expression.to_address())
        raise ValueError(f"{memory_reference_str} is not a memory reference")

    def _to_rs_memory_reference(self) -> quil_rs.MemoryReference:
        return quil_rs.MemoryReference(self.name, self.offset)

    def out(self) -> str:
        """Return the memory reference as a valid Quil string."""
        if self.declared_size is not None and self.declared_size == 1 and self.offset == 0:
            return f"{self.name}"
        else:
            return f"{self.name}[{self.offset}]"

    def __str__(self) -> str:
        if self.declared_size is not None and self.declared_size == 1 and self.offset == 0:
            return f"{self.name}"
        else:
            return f"{self.name}[{self.offset}]"

    def __repr__(self) -> str:
        return f"<MRef {self.name}[{self.offset}]>"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, MemoryReference) and other.name == self.name and other.offset == self.offset

    def __hash__(self) -> int:
        return hash((self.name, self.offset))

    def __getitem__(self, offset: int) -> "MemoryReference":
        if self.offset != 0:
            raise ValueError("Please only index off of the base MemoryReference (offset = 0)")

        # NOTE If a MemoryReference is the result of parsing (not
        # manually instantiated), it will likely be instantiated
        # without a declared_size, and so bounds checking will be
        # impossible.
        if self.declared_size and offset >= self.declared_size:
            raise IndexError("MemoryReference index out of range")

        return MemoryReference(name=self.name, offset=offset)

    def _substitute(
        self, d: ParameterSubstitutionsMapDesignator
    ) -> Union["MemoryReference", ExpressionValueDesignator]:
        if self in d:
            return d[self]

        return self


def _contained_mrefs(expression: ExpressionDesignator) -> set[MemoryReference]:
    """Determine which memory references are contained in this expression.

    :param expression: expression involving parameters
    :return: set of parameters contained in this expression
    """
    if isinstance(expression, BinaryExp):
        return _contained_mrefs(expression.op1) | _contained_mrefs(expression.op2)
    elif isinstance(expression, Function):
        return _contained_mrefs(expression.expression)
    elif isinstance(expression, MemoryReference):
        return {expression}
    else:
        return set()


class Frame(quil_rs.FrameIdentifier):
    """Representation of a frame descriptor."""

    def __new__(cls, qubits: Sequence[QubitDesignator], name: str) -> Self:
        """Initialize a new Frame."""
        return super().__new__(cls, name, _convert_to_rs_qubits(qubits))

    @classmethod
    def _from_rs_frame_identifier(cls, frame: quil_rs.FrameIdentifier) -> "Frame":
        return super().__new__(cls, frame.name, frame.qubits)

    @property  # type: ignore[override]
    def qubits(self) -> tuple[QubitDesignator, ...]:
        """Get the qubits in the frame."""
        return tuple(_convert_to_py_qubits(super().qubits))

    @qubits.setter
    def qubits(self, qubits: tuple[Qubit, FormalArgument]) -> None:
        quil_rs.FrameIdentifier.qubits.__set__(self, _convert_to_rs_qubits(qubits))  # type: ignore[attr-defined]

    def out(self) -> str:
        """Return the frame as a valid Quil string."""
        return super().to_quil()

    def __str__(self) -> str:
        """Return the frame as a string."""
        return super().to_quil_or_debug()


class WaveformInvocation(quil_rs.WaveformInvocation, QuilAtom):
    """A waveform invocation."""

    def __new__(cls, name: str, parameters: Optional[dict[str, ParameterDesignator]] = None) -> Self:
        """Initialize a new waveform invocation."""
        if parameters is None:
            parameters = {}
        rs_parameters = {key: _convert_to_rs_expression(value) for key, value in parameters.items()}
        return super().__new__(cls, name, rs_parameters)

    @property  # type: ignore[override]
    def parameters(self) -> dict[str, ParameterDesignator]:
        """The parameters in the waveform invocation."""
        return {key: _convert_to_py_expression(value) for key, value in super().parameters.items()}

    @parameters.setter
    def parameters(self, parameters: dict[str, ParameterDesignator]) -> None:
        rs_parameters = {key: _convert_to_rs_expression(value) for key, value in parameters.items()}
        quil_rs.WaveformInvocation.parameters.__set__(self, rs_parameters)  # type: ignore[attr-defined]

    def out(self) -> str:
        """Return the waveform invocation as a valid Quil string."""
        return self.to_quil()

    def __str__(self) -> str:
        """Return the frame as a string."""
        return super().to_quil_or_debug()


@deprecated(
    version="4.0",
    reason="The WaveformReference class will be removed, consider using WaveformInvocation instead.",
)
class WaveformReference(WaveformInvocation):
    """Representation of a Waveform reference."""

    def __new__(cls, name: str) -> Self:
        """Initialize a new waveform reference."""
        return super().__new__(cls, name, {})


def _template_waveform_property(
    name: str, *, dtype: Optional[Union[type[int], type[float]]] = None, doc: Optional[str] = None
) -> property:
    """Initialize a getters, setters, and deleter for a parameter on a ``TemplateWaveform``.

    Should only be used inside of ``TemplateWaveform`` or one its base classes.

    :param name: The name of the property
    :param dtype: An optional parameter that takes the int or float type, and attempts
            to convert the underlying complex value by casting the real part to `dtype`. If set, this function will
            raise an error if the complex number has a non-zero imaginary part.
    :param doc: Docstring for the property.
    """

    def fget(self: "TemplateWaveform") -> Optional[ParameterDesignator]:
        parameter = self.get_parameter(name)
        if parameter is None or dtype is None:
            return parameter

        if dtype is int or dtype is float:
            if isinstance(parameter, dtype):
                return parameter
            if not isinstance(parameter, complex):
                raise TypeError(
                    f"Requested float for parameter {name}, but a non-numeric value of type {type(parameter)} was "
                    "found instead"
                )
            if parameter.imag != 0.0:
                raise ValueError(
                    f"Requested float for parameter {name}, but a complex number with a non-zero imaginary part was "
                    "found"
                )
            return dtype(parameter.real)
        raise TypeError(f"TemplateWaveform is not compatible with dtype {dtype}")

    def fset(self: "TemplateWaveform", value: ParameterDesignator) -> None:
        self.set_parameter(name, value)

    def fdel(self: "TemplateWaveform") -> None:
        self.set_parameter(name, None)

    return property(fget, fset, fdel, doc)


class TemplateWaveform(quil_rs.WaveformInvocation, QuilAtom):
    """Base class for creating waveform templates."""

    NAME: ClassVar[str]

    def __new__(
        cls,
        name: str,
        *,
        duration: float,
        **kwargs: Union[Optional[ParameterDesignator], Optional[ExpressionDesignator]],
    ) -> Self:
        """Initialize a new TemplateWaveform."""
        rs_parameters = {key: _convert_to_rs_expression(value) for key, value in kwargs.items() if value is not None}
        rs_parameters["duration"] = _convert_to_rs_expression(duration)
        return super().__new__(cls, name, rs_parameters)

    def out(self) -> str:
        """Return the waveform as a valid Quil string."""
        return str(self)

    def get_parameter(self, name: str) -> Optional[ParameterDesignator]:
        """Get a parameter in the waveform by name."""
        parameter = super().parameters.get(name, None)
        if parameter is None:
            return None
        return _convert_to_py_expression(parameter)

    def set_parameter(self, name: str, value: Optional[ParameterDesignator]) -> None:
        """Set a parameter with a value."""
        parameters = super().parameters
        if value is None:
            parameters.pop(name, None)
        else:
            parameters[name] = _convert_to_rs_expression(value)
        quil_rs.WaveformInvocation.parameters.__set__(self, parameters)  # type: ignore[attr-defined]

    duration = _template_waveform_property("duration", dtype=float)

    def num_samples(self, rate: float) -> int:
        """Return the number of samples in the reference implementation of the waveform.

        Note: this does not include any hardware-enforced alignment (cf. documentation for `samples`).

        :param rate: The sample rate, in Hz.
        :return: The number of samples.
        """
        duration = self.duration.real
        if self.duration.imag != 0.0:
            raise ValueError("Can't calculate number of samples with a complex duration")
        return int(np.ceil(duration * rate))

    def samples(self, rate: float) -> np.ndarray:
        """Generate samples of the waveform.

        Note: this is close but not always exactly equivalent to the actual IQ
        values produced by the waveform generators on Rigetti hardware. The
        actual ADC process imposes some alignment constraints on the waveform
        duration (in particular, it must be compatible with the clock rate).

        :param rate: The sample rate, in Hz.
        :returns: An array of complex samples.

        """
        raise NotImplementedError()

    @classmethod
    def _from_rs_waveform_invocation(cls, waveform: quil_rs.WaveformInvocation) -> "TemplateWaveform":
        """Build a TemplateWaveform from a ``quil`` waveform invocation.

        The ``quil`` package has no equivalent to ``TemplateWaveform``s, this function checks the name and properties of
        a ``quil`` ``WaveformInvocation`` to see if they potentially match a subclass of ``TemplateWaveform``. If a
        match is found and construction succeeds, then that type is returned. Otherwise, a generic
        ``WaveformInvocation`` is returned.
        """
        from pyquil.quiltwaveforms import (
            BoxcarAveragerKernel,
            DragGaussianWaveform,
            ErfSquareWaveform,
            FlatWaveform,
            GaussianWaveform,
            HrmGaussianWaveform,
        )

        template: type[TemplateWaveform]  # mypy needs a type annotation here to understand this.
        for template in [
            FlatWaveform,
            GaussianWaveform,
            DragGaussianWaveform,
            HrmGaussianWaveform,
            ErfSquareWaveform,
            BoxcarAveragerKernel,
        ]:
            if template.NAME != waveform.name:
                continue
            parameter_names = [
                parameter[0] for parameter in inspect.getmembers(template, lambda a: isinstance(a, property))
            ]
            if set(waveform.parameters.keys()).issubset(parameter_names):
                try:
                    parameters = {key: _convert_to_py_expression(value) for key, value in waveform.parameters.items()}
                    return template(**parameters)  # type: ignore[arg-type]
                except TypeError:
                    break

        return super().__new__(cls, waveform.name, waveform.parameters)

    def __str__(self) -> str:
        return super().to_quil_or_debug()


def _update_envelope(
    iqs: np.ndarray,
    rate: float,
    scale: Optional[float],
    phase: Optional[float],
    detuning: Optional[float],
) -> np.ndarray:
    """Update a pulse envelope by optional shape parameters.

    The optional parameters are: 'scale', 'phase', 'detuning'.

    :param iqs: The basic pulse envelope.
    :param rate: The sample rate (in Hz).
    :return: The updated pulse envelope.
    """

    def default(obj: Optional[float], val: float) -> float:
        return obj if obj is not None else val

    scale = default(scale, 1.0)
    phase = default(phase, 0.0)
    detuning = default(detuning, 0.0)

    iqs *= scale * np.exp(1j * phase) * np.exp(1j * 2 * np.pi * detuning * np.arange(len(iqs)) / rate)

    return iqs


Waveform = Union[WaveformReference, TemplateWaveform]


def _convert_to_py_waveform(waveform: quil_rs.WaveformInvocation) -> Waveform:
    if not isinstance(waveform, quil_rs.WaveformInvocation):
        raise TypeError(f"{type(waveform)} is not a WaveformInvocation")
    if len(waveform.parameters) == 0:
        return WaveformReference(waveform.name)

    return TemplateWaveform._from_rs_waveform_invocation(waveform)



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""Sub-package for facilitating connections to the QVM / QPU."""

from qcs_sdk import QCSClient, RegisterMatrixConversionError
from qcs_sdk.qpu import RawQPUReadoutData
from qcs_sdk.qpu.api import ConnectionStrategy, ExecutionOptions, ExecutionOptionsBuilder
from qcs_sdk.qvm import RawQVMReadoutData

from pyquil.api._benchmark import BenchmarkConnection
from pyquil.api._compiler import (
    AbstractCompiler,
    EncryptedProgram,
    QPUCompiler,
    QPUCompilerAPIOptions,
    QuantumExecutable,
    QVMCompiler,
)
from pyquil.api._qam import QAM, MemoryMap, QAMExecutionResult
from pyquil.api._qpu import QPU, QPUExecuteResponse
from pyquil.api._quantum_computer import (
    QuantumComputer,
    get_qc,
    list_quantum_computers,
    local_forest_runtime,
)
from pyquil.api._qvm import QVM
from pyquil.api._wavefunction_simulator import WavefunctionSimulator
from pyquil.quantum_processor import QCSQuantumProcessor

__all__ = [
    "QCSClient",
    "RegisterMatrixConversionError",
    "RawQPUReadoutData",
    "RawQVMReadoutData",
    "ConnectionStrategy",
    "ExecutionOptions",
    "BenchmarkConnection",
    "AbstractCompiler",
    "EncryptedProgram",
    "QPUCompiler",
    "QPUCompilerAPIOptions",
    "QuantumExecutable",
    "QVMCompiler",
    "QAM",
    "MemoryMap",
    "QAMExecutionResult",
    "QPU",
    "QPUExecuteResponse",
    "QuantumComputer",
    "get_qc",
    "list_quantum_computers",
    "local_forest_runtime",
    "QVM",
    "WavefunctionSimulator",
    "QCSQuantumProcessor",
    "ExecutionOptionsBuilder",
]
